"""
    Baseline.BaselineUtils.py
"""
from importlib import reload
from molass.PackageUtils.NumbaUtils import get_ready_for_numba
get_ready_for_numba()
from pybaselines import Baseline

def molass_lpm_impl(x, y, **kwargs):
    """Linear plus minimum baseline implementation.

    Parameters
    ----------
    x : array-like
        The x-values of the data.
    y : array-like
        The y-values of the data.
    kwargs : dict
        Additional keyword arguments to pass to the baseline computation function.

    Returns
    -------
    baseline : array-like
        The computed baseline values.
    """
    from molass.Baseline.LpmBaseline import compute_lpm_baseline
    return compute_lpm_baseline(x, y, **kwargs)

def molass_uvdiff_impl(x, y, **kwargs):
    """UV difference baseline implementation.

    Parameters
    ----------
    x : array-like
        The x-values of the data.
    y : array-like
        The y-values of the data.
    kwargs : dict
        Additional keyword arguments to pass to the baseline computation function.

    Returns
    -------
    baseline : array-like
        The computed baseline values.
    """
    import molass.Baseline.UvdiffBaseline
    reload(molass.Baseline.UvdiffBaseline)  
    from molass.Baseline.UvdiffBaseline import compute_uvdiff_baseline
    return compute_uvdiff_baseline(x, y, **kwargs)

def molass_integral_impl(x, y, **kwargs):
    """Integral baseline implementation.

    Parameters
    ----------
    x : array-like
        The x-values of the data.
    y : array-like
        The y-values of the data.
    kwargs : dict
        Additional keyword arguments to pass to the baseline computation function.

    Returns
    -------
    baseline : array-like
        The computed baseline values.
    """
    from molass.Baseline.IntegralBaseline import compute_integral_baseline
    return compute_integral_baseline(x, y)

def pybaselines_asls_impl(x, y, **kwargs):
    """Asymmetric least squares baseline implementation using pybaselines.

    Parameters
    ----------
    x : array-like
        The x-values of the data.
    y : array-like
        The y-values of the data.
    kwargs : dict
        Additional keyword arguments to pass to the baseline computation function.

    Returns
    -------
    baseline : array-like
        The computed baseline values.
    """
    baseline_fitter = Baseline(x_data=x)
    baseline = baseline_fitter.asls(y, lam=1e7, p=0.02)[0]
    return baseline

def pybaselines_imor_impl(x, y, **kwargs):
    """Improved modified polynomial fitting baseline implementation using pybaselines.

    Parameters
    ----------
    x : array-like
        The x-values of the data.
    y : array-like
        The y-values of the data.
    kwargs : dict
        Additional keyword arguments to pass to the baseline computation function.

    Returns
    -------
    baseline : array-like
        The computed baseline values.
    """
    baseline_fitter = Baseline(x_data=x)
    baseline = baseline_fitter.imor(y, 10)[0]
    return baseline

def pybaselines_mormol_impl(x, y, **kwargs):
    """Morphological operations baseline implementation using pybaselines.
    Parameters
    ----------
    x : array-like
        The x-values of the data.
    y : array-like
        The y-values of the data.
    kwargs : dict
        Additional keyword arguments to pass to the baseline computation function.
    Returns
    -------
    baseline : array-like
        The computed baseline values.
    """
    baseline_fitter = Baseline(x_data=x)
    half_window = 100
    baseline = baseline_fitter.mormol(y, half_window, smooth_half_window=10, pad_kwargs={'extrapolate_window': 20})[0]
    return baseline

METHOD_DICT = {
    'linear': molass_lpm_impl,
    'uvdiff': molass_uvdiff_impl,
    'integral': molass_integral_impl,
    'asls': pybaselines_asls_impl,
    'imor': pybaselines_imor_impl,
    'mormol': pybaselines_mormol_impl,
}

def iterlen(a):
    """Get the length of an iterable, or 1 if it is not iterable.

    Parameters
    ----------
    a : any
        The object to check.

    Returns
    -------
    length : int
        The length of the iterable, or 1 if it is not iterable.
    """
    if type(a) is str:
        return 1
    try:
        return len(a)
    except:
        return 1

def get_baseline_func(method):
    """Get the baseline function(s) corresponding to the given method(s).

    Parameters
    ----------
    method : str, callable, or list/tuple of str/callable
        The method(s) to get the baseline function(s) for.
        If a single method is given, it will be used for both UV and XR baselines.
        If a list or tuple of two methods is given, the first will be used for UV and the second for XR.
        Each method can be either a string (key in METHOD_DICT) or a callable (custom function).

    Returns
    -------
    functions : list of callables
        A list containing the baseline function for UV and XR, respectively.
    """
    if method is None:
        method = 'linear'
    num_methods = iterlen(method)
    if num_methods == 1:
        methods = [method, method]
    elif num_methods == 2:
        methods = method
    else:
        raise TypeError(f"given number of methods {num_methods} != 2")

    ret_methods = []
    for m in methods:
        if type(m) is str:
            func = METHOD_DICT[m]
        elif callable(m):
            func = m
        else:
            raise TypeError(f"method should be either str type, callable, or a pair of those")
        ret_methods.append(func)
    return ret_methods

def get_uv_baseline_func(method):
    """Get the UV baseline function corresponding to the given method.

    Parameters
    ----------
    method : str, callable, or list/tuple of str/callable
        The method(s) to get the baseline function(s) for.
        If a single method is given, it will be used for both UV and XR baselines.
        If a list or tuple of two methods is given, the first will be used for UV and the second for XR.
        Each method can be either a string (key in METHOD_DICT) or a callable (custom function).

    Returns
    -------
    function : callable
        The baseline function for UV.
    """
    return get_baseline_func(method)[0]

def get_xr_baseline_func(method):
    """Get the XR baseline function corresponding to the given method.
    
    Parameters
    ----------
    method : str, callable, or list/tuple of str/callable
        The method(s) to get the baseline function(s) for.
        If a single method is given, it will be used for both UV and XR baselines.
        If a list or tuple of two methods is given, the first will be used for UV and the second for XR.
        Each method can be either a string (key in METHOD_DICT) or a callable (custom function).

    Returns
    -------
    function : callable
        The baseline function for XR.
    """
    return get_baseline_func(method)[1]


"""
    Arrow3D.py

    borrowed from https://gist.github.com/WetHat/1d6cd0f7309535311a539b42cccca89c
"""
import numpy as np
from matplotlib.patches import FancyArrowPatch
from mpl_toolkits.mplot3d.proj3d import proj_transform
from mpl_toolkits.mplot3d.axes3d import Axes3D

class Arrow3D(FancyArrowPatch):

    def __init__(self, x, y, z, dx, dy, dz, *args, **kwargs):
        super().__init__((0, 0), (0, 0), *args, **kwargs)
        self._xyz = (x, y, z)
        self._dxdydz = (dx, dy, dz)

    def draw(self, renderer):
        x1, y1, z1 = self._xyz
        dx, dy, dz = self._dxdydz
        x2, y2, z2 = (x1 + dx, y1 + dy, z1 + dz)

        xs, ys, zs = proj_transform((x1, x2), (y1, y2), (z1, z2), self.axes.M)
        self.set_positions((xs[0], ys[0]), (xs[1], ys[1]))
        super().draw(renderer)
        
    def do_3d_projection(self, renderer=None):
        x1, y1, z1 = self._xyz
        dx, dy, dz = self._dxdydz
        x2, y2, z2 = (x1 + dx, y1 + dy, z1 + dz)

        xs, ys, zs = proj_transform((x1, x2), (y1, y2), (z1, z2), self.axes.M)
        self.set_positions((xs[0], ys[0]), (xs[1], ys[1]))

        return np.min(zs) 

def _arrow3D(ax, x, y, z, dx, dy, dz, *args, **kwargs):
    '''Add an 3d arrow to an `Axes3D` instance.'''

    arrow = Arrow3D(x, y, z, dx, dy, dz, *args, **kwargs)
    ax.add_artist(arrow)

def add_arrow3D():
    """
    Add annotate3D method to Axes3D class.
    """
    setattr(Axes3D, 'arrow3D', _arrow3D)


"""
    Baseline.Basesurface.py
"""
import numpy as np
from molass.Baseline.Surface import Surface

def get_linear_surface(icurve, jcurve):
    """Get a linear surface from the given intensity curves.
    
    Parameters
    ----------
    icurve : IntensityCurve
        The intensity curve along the i-axis.
    jcurve : IntensityCurve
        The intensity curve along the j-axis.
        
    Returns
    -------
    surface : 2D array-like
        The linear surface defined by the outer product of the two curves.
    """
    return icurve.y[:,np.newaxis] @ jcurve.y[np.newaxis,:]



"""
    Stats.Moment.py

    Copyright (c) 2024-2025, SAXS Team, KEK-PF
"""
import numpy as np

def compute_meanstd(x, y):
    """Compute the mean and standard deviation of a distribution defined by x and y.

    Parameters
    ----------
    x : array-like
        The x-values of the distribution.
    y : array-like
        The y-values of the distribution.

    Returns
    -------
    M1 : float
        The mean of the distribution.
    std : float
        The standard deviation of the distribution.
    """
    W = np.sum(y)
    M1 = np.sum(x*y)/W              # raw moment
    M2 = np.sum(y*(x-M1)**2)/W      # central moment
    return M1, np.sqrt(M2)

class Moment:
    """A class to represent the moment of a distribution defined by x and y.
    Attributes
    ----------
    x : array-like
        The x-values of the distribution.
    y : array-like
        The y-values of the distribution.
    y_ : array-like or None
        The processed y-values of the distribution. If None, it has not been computed yet.
    M : float or None
        The mean of the distribution. If None, it has not been computed yet.
    std : float or None
        The standard deviation of the distribution. If None, it has not been computed yet.
    lpm_percent : float or None
        The estimated percentage of low-q plateau in the distribution. If None, it has not been computed yet.
    """
    def __init__(self, x, y):
        """Initialize the Moment object.
        Parameters
        ----------
        x : array-like
            The x-values of the distribution.
        y : array-like
            The y-values of the distribution.
        """
        self.x = x
        self.y = y
        self.y_ = None
        self.M = None       # to avoid repeated computation in case with multiple references
        self.std = None
        self.lpm_percent = None
    
    def get_y_(self, **kwargs):
        """Get the processed y-values of the distribution.
            If y_ is None, it will be computed using the provided processing function.

        Parameters
        ----------
        **kwargs : keyword arguments
            Additional keyword arguments to pass to the processing function.

        Returns
        -------
        y_ : array-like
            The processed y-values of the distribution.
        """
        if self.y_ is None:
            y_ = self.y.copy()
            y_[y_ < 0] = 0  # negative values are usually inappropriate for weights
            self.y_ = y_
        return self.y_
    
    def debug_plot(self, ax):
        """Plot the original and processed y-values for debugging purposes.
        
        Parameters
        ----------
        ax : matplotlib.axes.Axes
            The axes to plot on.
        """
        y_ = self.get_y_()
        ax.plot(self.x, self.y, label='y')
        ax.plot(self.x, y_, ":", label='y_')
        ax.legend()
 
    def get_meanstd(self):
        """Get the mean and standard deviation of the distribution.
            If M and std are None, they will be computed using the processed y-values.

        Returns
        -------
        M : float
            The mean of the distribution.
        std : float
            The standard deviation of the distribution.
        """
        if self.y_ is None:
            y_ = self.get_y_()
            self.M, self.std = compute_meanstd(self.x, y_)
        return self.M, self.std
    
    def is_in_nsigma(self, n, px):
        """Check if a given x-value is within n standard deviations from the mean.
        
        Parameters
        ----------
        n : float
            The number of standard deviations.
        px : float
            The x-value to check.

        Returns
        -------
        bool
            True if px is within n standard deviations from the mean, False otherwise.
        """
        M, std = self.get_meanstd()
        return M - n*std < px and px < M + n*std
    
    def get_nsigma_points(self, n):
        """Get the indices of the x-values that are within n standard deviations from the mean.

        Parameters
        ----------
        n : float
            The number of standard deviations.

        Returns
        -------
        (i, j) : tuple of int
            The indices of the first and last x-values within n standard deviations from the mean.
            If no such points exist, returns (None, None).
        """
        M, std = self.get_meanstd()
        try:
            wanted_range = np.logical_and(M - n*std < self.x, self.x < M + n*std)
            i, j = np.where(wanted_range)[0][[0,-1]]
            return i, j
        except Exception as e:
            print(f"Error in get_nsigma_points: {e}")
            import matplotlib.pyplot as plt
            print("M, std, n, y_:", M, std, n, self.y_)
            plt.plot(self.x, self.y, label='y')
            plt.plot(self.x, self.y_, label='y_')
            plt.axvline(M - n*std, color='r', linestyle='--', label='Lower Bound')
            plt.axvline(M + n*std, color='r', linestyle='--', label='Upper Bound')
            plt.legend()
            plt.show()
            return None, None

    def get_lpm_percent(self, debug=True):
        """Estimate the percentage of low-q plateau in the distribution.
        
        Parameters
        ----------
        debug : bool, optional
            If True, enables debug mode for more verbose output, by default True.

        Returns
        -------
        lpm_percent : float
            The estimated percentage of low-q plateau in the distribution.
        """
        if self.lpm_percent is None or debug:
            if debug:
                from importlib import reload
                import molass.Baseline.LpmBaseline
                reload(molass.Baseline.LpmBaseline)
            from molass.Baseline.LpmBaseline import estimate_lpm_percent
            self.lpm_percent = estimate_lpm_percent(self)
        return self.lpm_percent



import os
import sys
import importlib.util
# --- Prepend pythonpath from PYTHONPATH env if set, or from pyproject.toml ---
from tools.pyproject_utils import get_pythonpath_from_pyproject
pythonpath_env = os.environ.get('PYTHONPATH', '')
if pythonpath_env:
    for p in reversed(pythonpath_env.split(os.pathsep)):
        if p and p not in sys.path:
            sys.path.insert(0, p)
else:
    # Fallback: use pyproject.toml if present (for direct runs)
    for abs_p in reversed(get_pythonpath_from_pyproject()):
        if abs_p not in sys.path:
            sys.path.insert(0, abs_p)
sys.path.insert(0, '.')     # Ensure current directory is in sys.path


# Support node id: test_file.py::test_func
if len(sys.argv) < 2:
    print("Usage: python interactive_test_script.py <test_file.py>[::test_func]")
    sys.exit(1)

node_id = sys.argv[1]
if '::' in node_id:
    file_part, func_part = node_id.split('::', 1)
else:
    file_part, func_part = node_id, None
abs_test_path = os.path.abspath(file_part)

# Set environment variables (optional: could be passed from parent)
os.environ.setdefault('MOLASS_ENABLE_PLOTS', 'true')
os.environ.setdefault('MOLASS_SAVE_PLOTS', 'false')
os.environ.setdefault('MOLASS_PLOT_DIR', 'test_plots')

print(f"Python executable: {sys.executable}")
print(f"Current working directory: {os.getcwd()}")
print(f"Test file path: {abs_test_path}")
print(f"Test file exists: {os.path.exists(abs_test_path)}")

try:
    spec = importlib.util.spec_from_file_location("test_module", abs_test_path)
    test_module = importlib.util.module_from_spec(spec)
    print(f"Module spec created successfully")
    spec.loader.exec_module(test_module)
    print(f"Module loaded successfully")
except Exception as e:
    print(f"Error loading module: {e}")
    import traceback
    traceback.print_exc()
    test_module = None


# Find all test functions (including decorated ones)
test_functions = []
if test_module:
    print(f"Module attributes: {[name for name in dir(test_module) if not name.startswith('__')]}")
    for name in dir(test_module):
        obj = getattr(test_module, name)
        if name.startswith('test_') and callable(obj):
            test_functions.append((name, obj))
            print(f"Found test function: {name}")
else:
    print("No test module loaded, cannot find test functions")

# Sort test functions by name for predictable order
test_functions.sort(key=lambda x: x[0])

if func_part:
    # Only run the specified function (exact match)
    selected = [(name, func) for name, func in test_functions if name == func_part]
    if not selected:
        print(f"No test function named '{func_part}' found in module.")
    else:
        print(f"Running only test function: {func_part}")
    test_functions = selected

print(f"Found {len(test_functions)} test functions: {[name for name, func in test_functions]}")

# Run each test function
for test_name, test_func in test_functions:
    print(f"\nRunning {test_name}...")
    try:
        test_func()
        print(f"✅ {test_name} PASSED")
    except Exception as e:
        print(f"❌ {test_name} FAILED: {e}")
        import traceback
        traceback.print_exc()

print("\nAll test functions completed.")

# Show plots if interactive mode is enabled and matplotlib.pyplot is imported
if os.environ.get('MOLASS_ENABLE_PLOTS', 'false').lower() == 'true':
    print("[DEBUG] Entering plot display block at end of script.")
    try:
        import matplotlib
        matplotlib.use('TkAgg')
        import matplotlib.pyplot as plt
        fignums = plt.get_fignums()
        print(f"[DEBUG] Number of open figures: {len(fignums)}; Figure numbers: {fignums}")
        print("Calling plt.show() to display plots (interactive mode)...")
        plt.show()
    except ImportError:
        print("matplotlib is not installed; skipping plt.show()")
    except Exception as e:
        print(f"Error calling plt.show(): {e}")



"""
    Batch.V1Result.py

    This module contains the class V1Result, which is used to store the results of the V1 batch process.

    Copyright (c) 2025, SAXS Team, KEK-PF 
"""

class V1Result:
    """
    A class to represent the results of the V1 batch process.
    """

    def __init__(self):
        """
        """
        pass


"""
    DataUtils.XrLoader.py
"""
from glob import glob
import numpy as np

def load_xr(folder_path):
    """
    Load X-ray scattering data from a folder containing .dat files.

    Parameters
    ----------
    folder_path : str
        Path to the folder containing .dat files.

    Returns
    -------
    xr_array : np.ndarray
        3D array containing the X-ray scattering data.

    datafiles : list of str
        List of data file paths corresponding to the loaded data.

    Notes
    -----
    The function assumes that each .dat file contains data in a format compatible with np.loadtxt.
    The first dimension corresponds to the number of files, the second to the number of points, and the third to the data columns.
    """
    input_list = []
    datafiles = []
    for path in sorted(glob(folder_path + "/*.dat")):
        try:
            input_list.append(np.loadtxt(path))
            datafiles.append(path)
        except Exception as e:
            print(f"Error loading {path}: {e}")
    try:
        xr_array = np.array(input_list)
    except ValueError as e:
        print(f"Error converting input list to array: {e}")
        from molass_legacy.SerialAnalyzer.SerialDataUtils import convert_to_the_least_shape
        xr_array = np.array(convert_to_the_least_shape(input_list)[0])
        print(f"Converted to least shape array with shape {xr_array.shape}")
    except Exception:
        raise
    return xr_array, datafiles

def xr_remove_bubbles(xr_array, logger=None, debug=False):
    """Remove bubbles from the XR data array.

    Parameters
    ----------
    xr_array : np.ndarray
        3D array containing the X-ray scattering data.
    logger : logging.Logger, optional
        Logger for logging messages. If None, print to console.
    debug : bool, optional
        If True, enable debug mode with additional output.
    """
    from molass.DataObjects.Curve import create_icurve
    from molass.DataUtils.AnomalyHandlers import bubble_check_impl, remove_bubbles_impl
    qv = xr_array[0,:,0]
    xrM = xr_array[:,:,1].T
    x = np.arange(xrM.shape[1])
    icurve = create_icurve(x, xrM, qv, pickvalue=0.02)
    bubbles = bubble_check_impl(icurve.y)

    if debug:
        from copy import deepcopy
        print(f"bubbles: {bubbles}")
        icurve_orig = deepcopy(icurve)

    if len(bubbles) > 0:
        excluded_set = set()
        remove_bubbles_impl(xr_array, bubbles, excluded_set)
        if logger is None:
            print("bubbles have been removed at %s" % bubbles)
        else:
            logger.warning("bubbles have been removed at %s", bubbles)

    if debug:
        import matplotlib.pyplot as plt
        xrM_ = xr_array[:,:,1].T
        icurve_ = create_icurve(x, xrM_, qv, pickvalue=0.02)
        fig, (ax1, ax2) = plt.subplots(ncols=2, figsize=(12, 5))
        ax1.plot(*icurve_orig.get_xy())
        ax2.plot(*icurve_.get_xy())
        plt.show()

def load_xr_with_options(folder_path, remove_bubbles=False, logger=None, debug=False):
    """Load X-ray scattering data from a folder with options to preprocess.

    Parameters
    ----------
    folder_path : str
        Path to the folder containing .dat files.
    remove_bubbles : bool, optional
        If True, remove bubbles from the data, by default False.
    logger : logging.Logger, optional
        Logger for logging messages. If None, print to console.
    debug : bool, optional
        If True, enable debug mode with additional output, by default False.
        
    Returns
    -------
    xr_array : np.ndarray
        3D array containing the X-ray scattering data.
    datafiles : list of str
        List of data file paths corresponding to the loaded data.
    """
    xr_array, datafiles = load_xr(folder_path)
    if remove_bubbles:
        xr_remove_bubbles(xr_array, logger=logger, debug=debug)
    return xr_array, datafiles


"""
Simple test of trajectory tracking without animation rendering.
"""

import sys
import numpy as np
import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import Axes3D
from pathlib import Path

molass_path = Path(__file__).parent.parent
sys.path.insert(0, str(molass_path))

from molass.SEC.ColumnSimulation import get_animation

print("Running simple trajectory test...")
print("Tracking particle 1250 for 200 frames...")

# Get animation with trajectory tracking
anim, stats = get_animation(
    num_frames=200,
    seed=42,
    close_plot=True,
    track_particle_id=1250,
    use_tqdm=True,
    blit=False
)

print("\nTrajectory data collected!")
trajectory = stats['trajectory']

print(f"Particle type: {trajectory['particle_type']}")  # 2 = small (red)
print(f"Number of frames: {len(trajectory['positions'])}")
print(f"Final cumulative adsorbed time: {trajectory['cumulative_adsorbed_time'][-1]:.6f}")

# Quick 3D plot
fig = plt.figure(figsize=(10, 8))
ax = fig.add_subplot(111, projection='3d')

positions = trajectory['positions']
cum_time = trajectory['cumulative_adsorbed_time']

x = positions[:, 0]
y = positions[:, 1]
z = cum_time

# Plot trajectory
ax.plot(x, y, z, 'k-', linewidth=1, alpha=0.7)
ax.scatter([x[0]], [y[0]], [z[0]], c='lime', s=100, marker='o', label='Start')
ax.scatter([x[-1]], [y[-1]], [z[-1]], c='red', s=100, marker='s', label='End')

ax.set_xlabel('X Position')
ax.set_ylabel('Y Position')
ax.set_zlabel('Cumulative Adsorbed Time (Lévy Process)')
ax.set_title('3D Trajectory - Compound Poisson Process')
ax.legend()

output_path = molass_path / 'study' / 'test_levy_trajectory_3d.png'
plt.savefig(output_path, dpi=150)
print(f"\nPlot saved to: {output_path}")

plt.show()



"""
Simple viewer for the generated comparison images
"""
import matplotlib.pyplot as plt
import matplotlib.image as mpimg

# Load images
img1 = mpimg.imread('comparison1_exponential.png')
img2 = mpimg.imread('comparison2_bimodal.png')
img3 = mpimg.imread('comparison3_various_shapes.png')

# Display in separate windows
fig1 = plt.figure(figsize=(10, 8))
plt.imshow(img1)
plt.axis('off')
plt.title('Comparison 1: Level (a) vs Level (b) with Exponential Distribution', fontsize=14)
plt.tight_layout()

fig2 = plt.figure(figsize=(10, 11))
plt.imshow(img2)
plt.axis('off')
plt.title('Comparison 2: Level (b) with Bimodal Distribution', fontsize=14)
plt.tight_layout()

fig3 = plt.figure(figsize=(10, 8))
plt.imshow(img3)
plt.axis('off')
plt.title('Comparison 3: Effect of Different Distribution Shapes', fontsize=14)
plt.tight_layout()

plt.show()



"""
    DataUtils.PfBeamline.py
"""
import os

BEAMLINE_NAME = {
    "QEPB0040" : "PF BL-10C",
    "QEP01069" : "PF BL-15A2",
    "QEP03487" : "SPring-8 BL38B1",
    "FLMS00687" : "SPring-8 BL38B1",
    }

class BeamlineInfo:
    """
    A class to hold information about a beamline.
    """
    def __init__(self, **entries): 
        """Initialize the BeamlineInfo object.

        Parameters
        ----------
        **entries : dict
            Keyword arguments to initialize the beamline information.
        """
        self.__dict__.update(entries)

    def get_concfactor(self):
        """
        Get the concentration factor, which is the product of path length and extinction.

        Returns
        -------
        float
            The concentration factor.
        """
        return self.path_length * self.extinction

    def __str__(self):
        return str(self.__dict__)

def get_beamlineinfo_from_folder(folder):
    """
    Get beamline information from a folder containing UV data files.

    Parameters
    ----------
    folder : str
        The folder path containing the UV data files.

    Returns
    -------
    BeamlineInfo
        The beamline information object."""
    from glob import glob
    from molass_legacy.SerialAnalyzer.SerialDataUtils import get_uv_filename

    uv_device_id = None
    beamline = None
    file = get_uv_filename(folder)
    if file is not None:
        uvfile_path = os.path.join(folder, file)
        with open(uvfile_path) as fh:
            for k, line in enumerate(fh):
                if line.find('Spectrometers') >= 0:
                    uv_device_id = line.strip().split()[-1]
                    beamline = BEAMLINE_NAME.get(uv_device_id)
                    break

    return BeamlineInfo(uv_device_id=uv_device_id, name=beamline)

def get_beamlineinfo_from_settings():
    """
    Get beamline information from settings.
    
    Returns
    -------
    BeamlineInfo
        The beamline information object.
    """
    from molass_legacy._MOLASS.SerialSettings import get_setting
    uv_device_id = get_setting("uv_device_no")
    beamline = BEAMLINE_NAME.get(uv_device_id)
    path_length = get_setting('path_length')
    extinction = get_setting('extinction')
    return BeamlineInfo(uv_device_id=uv_device_id,
                        name=beamline,
                        path_length=path_length,
                        extinction=extinction,
                        )


"""
    test Baseline Correction
"""
import os
import sys
repo_root = os.path.abspath(os.path.join(os.path.dirname(__file__), '..', '..', '..'))
sys.path.insert(0, repo_root)
import matplotlib.pyplot as plt
from molass import get_version
get_version(toml_only=True)     # to ensure that the current repository is used
from molass.Local import get_local_settings
local_settings = get_local_settings()
DATA_ROOT_FOLDER = local_settings['DATA_ROOT_FOLDER']
from molass.Testing import control_matplotlib_plot, is_interactive

@control_matplotlib_plot
def test_010_OA_Ald_default():
    from molass_data import SAMPLE1
    from molass.DataObjects import SecSaxsData as SSD
    ssd = SSD(SAMPLE1)
    ssd.plot_compact(baseline=True, debug=is_interactive())
    trimmed_ssd = ssd.trimmed_copy()
    corrected_ssd = trimmed_ssd.corrected_copy(debug=is_interactive())
    corrected_ssd.plot_compact(baseline=True, debug=is_interactive())

@control_matplotlib_plot
def test_020_OA_Ald_uvdiff():
    from molass_data import SAMPLE1
    from molass.DataObjects import SecSaxsData as SSD
    ssd = SSD(SAMPLE1)
    ssd.set_baseline_method(('linear', 'uvdiff'))
    ssd.plot_compact(baseline=True, debug=is_interactive())
    trimmed_ssd = ssd.trimmed_copy()
    corrected_ssd = trimmed_ssd.corrected_copy(debug=is_interactive())
    corrected_ssd.plot_compact(baseline=True, debug=is_interactive())

@control_matplotlib_plot
def test_031_OA_Ald_integral():
    from molass_data import SAMPLE1
    from molass.DataObjects import SecSaxsData as SSD
    ssd = SSD(SAMPLE1)
    ssd.set_baseline_method('integral')
    ssd.plot_compact(baseline=True, debug=is_interactive())
    trimmed_ssd = ssd.trimmed_copy()
    corrected_ssd = trimmed_ssd.corrected_copy(debug=is_interactive())
    corrected_ssd.plot_compact(baseline=True, debug=is_interactive())

@control_matplotlib_plot
def test_032_SAMPLE2_integral():
    from molass_data import SAMPLE2
    from molass.DataObjects import SecSaxsData as SSD
    ssd = SSD(SAMPLE2)
    ssd.set_baseline_method('integral')
    ssd.plot_compact(baseline=True, debug=is_interactive())
    trimmed_ssd = ssd.trimmed_copy()
    corrected_ssd = trimmed_ssd.corrected_copy(debug=is_interactive())
    corrected_ssd.plot_compact(baseline=True, debug=is_interactive())

def test_040_endpoint_fraction_differs_from_lpm():
    """get_baseline2d(endpoint_fraction=0.15) must differ from standard LPM.

    Uses a synthetic SsMatrixData-like object with a planted negative dip so
    that the standard LPM anchor is contaminated, then verifies that the
    endpoint-anchored baseline is numerically different and that the standard
    path is unchanged (endpoint_fraction=None produces the original result).
    Also verifies new arg order (M, iv, jv) and E optional.
    """
    import numpy as np
    from molass.DataObjects.SsMatrixData import SsMatrixData

    rng = np.random.default_rng(42)
    n_q, n_frames = 10, 200
    frames = np.arange(n_frames, dtype=float)

    # True linear baseline: slope=0.005, intercept=1.0 per q-row
    true_bl = 0.005 * frames + 1.0
    # Gaussian protein peak + negative dip after it
    peak = 2.0 * np.exp(-0.5 * ((frames - 80) / 8) ** 2)
    neg_dip = -0.6 * np.exp(-0.5 * ((frames - 130) / 6) ** 2)
    signal = true_bl + peak + neg_dip
    M = np.tile(signal, (n_q, 1)) + rng.normal(0, 0.01, (n_q, n_frames))
    q = np.linspace(0.01, 0.3, n_q)

    # Issue #49: M first, then iv, jv
    # Issue #47: E omitted (defaults to None)
    ssd = SsMatrixData(M, q, frames)

    bl_standard = ssd.get_baseline2d(method='linear')
    bl_endpoint = ssd.get_baseline2d(method='linear', endpoint_fraction=0.15)

    # They must differ
    assert not np.allclose(bl_standard, bl_endpoint), \
        "endpoint baseline should differ from standard LPM"

    # Endpoint baseline should be closer to the true linear baseline
    err_standard = np.mean((bl_standard - true_bl) ** 2)
    err_endpoint = np.mean((bl_endpoint - true_bl) ** 2)
    assert err_endpoint < err_standard, \
        f"endpoint MSE ({err_endpoint:.6f}) should be < LPM MSE ({err_standard:.6f})"

    # Calling without endpoint_fraction still gives the original LPM result
    bl_default = ssd.get_baseline2d(method='linear')
    np.testing.assert_array_equal(bl_standard, bl_default)


if __name__ == "__main__":
    # test_010_OA_Ald_default()
    # test_020_OA_Ald_uvdiff()
    # test_031_OA_Ald_integral()
    test_032_SAMPLE2_integral()
    # plt.show()


"""
    PlotUtils.SecSaxsDataPlot.py

    Copyright (c) 2025, SAXS Team, KEK-PF
"""
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.gridspec import GridSpec
from molass.PlotUtils.MatrixPlot import simple_plot_3d

def plot_3d_impl(ssd, xr_only=False, uv_only=False, **kwargs):
    matrixplot_kwargs = {}
    title = kwargs.pop('title', None)
    with_2d_section_lines = kwargs.pop('with_2d_section_lines', False)
    if with_2d_section_lines:
        mapping = ssd.get_mapping()
        xr_curve = mapping.xr_curve
        uv_curve = mapping.uv_curve
        alpha = 0.4
    else:
        alpha = 1
    matrixplot_kwargs['cmap'] = 'coolwarm'
    matrixplot_kwargs['alpha'] = alpha
    matrixplot_kwargs['view_init'] = kwargs.get('view_init', {})
    matrixplot_kwargs['view_arrows'] = kwargs.get('view_arrows', False)
    wide_margin_layout = kwargs.get('wide_margin_layout', False)

    if xr_only or uv_only:
        ncols = 1
        figsize = (6,5)
    else:
        ncols = 2
        figsize = (12,5)
    
    if wide_margin_layout:
        fig = plt.figure(figsize=figsize)
        gs = GridSpec(1,9)
        axs0 = fig.add_subplot(gs[0, 0])
        axs4 = fig.add_subplot(gs[0, 4])
        axs8 = fig.add_subplot(gs[0, 8])
        for ax in [axs0, axs4, axs8]:
            ax.set_axis_off()
        axes = [fig.add_subplot(gs[0,1:4], projection='3d'), fig.add_subplot(gs[0,5:8], projection='3d')]
    else:
        fig, axes = plt.subplots(ncols=ncols, figsize=figsize, subplot_kw=dict(projection='3d'))

    if title is not None:
        fig.suptitle(title)

    labelkwarg = dict(fontsize=9)
    tickkwarg = dict(labelsize=9)

    if uv_only:
        ax1 = axes
        ax2 = None
    elif xr_only:
        ax1 = None
        ax2 = axes
    else:
        ax1, ax2 = axes

    if ax1 is not None:
        ax1.set_title("UV")
        uv = ssd.uv
        if uv is not None:
            ax1.set_xlabel("wavelength", **labelkwarg)
            ax1.set_ylabel("frames", **labelkwarg)
            ax1.set_zlabel("absorbance", **labelkwarg)
            simple_plot_3d(ax1, uv.M, x=uv.iv, y=uv.jv, **matrixplot_kwargs)
            for axis in [ax1.xaxis, ax1.yaxis, ax1.zaxis]:
                axis.set_tick_params(**tickkwarg)

        if with_2d_section_lines:   
            uv_x, uv_y = uv_curve.get_xy()
            ww = np.ones(len(uv_x)) * 280
            ax1.plot(ww, uv_x, uv_y, color='green', label="280 nm section line")

            m = xr_curve.get_max_i()
            n = mapping.get_mapped_index(m, xr_curve.x, uv_curve.x)
            uv_jcurve = ssd.uv.get_jcurve(j=n)
            ff = np.ones(len(uv_jcurve.x)) * n
            ax1.plot(uv_jcurve.x, ff, uv_jcurve.y, color="cyan", label="UV Absorbance at j=%d" % n)

            ax1.legend()

    if ax2 is not None:
        ax2.set_title("XR (X-ray)")
        xr = ssd.xr
        if xr is not None:
            ax2.set_xlabel("Q", **labelkwarg)
            ax2.set_ylabel("frames", **labelkwarg)
            ax2.set_zlabel("scattering", **labelkwarg)
            simple_plot_3d(ax2, xr.M, x=xr.iv, y=xr.jv, **matrixplot_kwargs)
            for axis in [ax2.xaxis, ax2.yaxis, ax2.zaxis]:
                axis.set_tick_params(**tickkwarg)

        if with_2d_section_lines:     
            xr_x, xr_y = xr_curve.get_xy()
            qq = np.ones(len(xr_x)) * 0.02
            ax2.plot(qq, xr_x, xr_y, color='green', label=r"0.02 $\AA^{-1}$ section line")
            xr_jcurve = ssd.xr.get_jcurve(j=m)
            ff = np.ones(len(xr_jcurve.x)) * m
            ax2.plot(xr_jcurve.x, ff, xr_jcurve.y, color="orange", alpha=0.5, label="XR Scattering at j=%d" % m)

            ax2.legend()

    fig.tight_layout()

    from molass.PlotUtils.PlotResult import PlotResult
    return PlotResult(fig, (ax1, ax2))

def plot_compact_impl(ssd, **kwargs):
    from molass.PlotUtils.TrimmingPlot import ij_from_slice
    from molass.PlotUtils.TwinAxesUtils import align_zero_y
    from molass.Global.Options import get_molass_options
    from molass.DataObjects.Curve import Curve

    debug = kwargs.get('debug', False)

    title = kwargs.pop('title', None)
    baseline = kwargs.pop('baseline', False)
    ratio_curve = kwargs.pop('ratio_curve', False)
    moment_lines = kwargs.pop('moment_lines', False)

    trim = ssd.make_trimming()
    mapping = ssd.get_mapping()
    uv_curve = mapping.uv_curve

    # Use the recognition curve for XR elution display (consistent with baseline)
    recognition = get_molass_options('elution_recognition')
    if recognition == 'sum':
        xr_curve = ssd.xr.get_recognition_curve()
        xr_label = "XR Elution (sum)"
    else:
        xr_curve = mapping.xr_curve
        xr_label = "XR Elution at Q=0.02"

    x = xr_curve.x
    mp_curve = mapping.get_mapped_curve(xr_curve, uv_curve, inverse_range=True, debug=debug)
    xr_max_x, xr_max_y = xr_curve.get_max_xy()
    
    fig = plt.figure(figsize=(12, 5))
    if title is not None:
        fig.suptitle(title)
    gs = GridSpec(2, 2)

    # Plot the UV and XR elution curves
    ax1 = fig.add_subplot(gs[:, 0])
    ax1.plot(x, xr_curve.y, color="orange", alpha=0.5, label=xr_label)
    axt = ax1.twinx()
    axt.grid(False)
    uv_wl = ssd.uv.pickat
    axt.plot(mp_curve.x, mp_curve.y, linestyle=":", color="C0", label=f"mapped UV Elution at wavelength={uv_wl}")

    if baseline:
        uv_baseline = ssd.uv.get_ibaseline(debug=False)
        if recognition == 'sum':
            # Compute 1D baseline for the recognition curve (fast)
            from molass.Baseline.BaselineUtils import get_xr_baseline_func
            method = ssd.xr.get_baseline_method()
            compute_bl = get_xr_baseline_func(method)
            bl_y = compute_bl(xr_curve.x, xr_curve.y, moment=ssd.xr.get_moment())
            xr_baseline = Curve(xr_curve.x, bl_y, type='i')
        else:
            xr_baseline = ssd.xr.get_ibaseline(debug=False)
        mp_baseline = mapping.get_mapped_curve(xr_baseline, uv_baseline, inverse_range=True, debug=False)
        ax1.plot(xr_baseline.x, xr_baseline.y, color='red', label="XR Baseline")
        axt.plot(mp_baseline.x, mp_baseline.y, ls=':', color='red', label="UV Baseline")
        axt.legend(loc="center left")

    align_zero_y(ax1, axt)

    ymin, ymax = ax1.get_ylim()
    ax1.set_ylim(ymin, ymax * 1.2)
    i, j = ij_from_slice(trim.xr_slices[1])
    ax1.axvspan(*x[[i,j]], color='green', alpha=0.1)
    ax1.axvline(xr_max_x, color='yellow')
    ax1.set_xlabel("Frame Index for X-ray")    

    if ratio_curve:
        axt = ax1.twinx()
        axt.grid(False)
        ratio_curve = mapping.compute_ratio_curve(mp_curve=mp_curve, debug=debug)
        axt.plot(*ratio_curve.get_xy(), color="C2", alpha=0.5, label="UV/XR Ratio")
        ymin, ymax = axt.get_ylim()
        axt.set_ylim(0, ymax * 1.2)
        axt.legend(loc="center left")

    if moment_lines:
        moment = xr_curve.get_moment()
        xmin, xmax = ax1.get_xlim()
        for n in [1, 2, 3]:
            i, j = moment.get_nsigma_points(n)
            xi = xr_curve.x[i]
            xj = xr_curve.x[j]
            label= r'XR n-$\sigma$ bounds' if n == 1 else None
            labeled = False
            if xi >= xmin:
                ax1.axvline(xi, color='gray', linestyle=':', alpha=0.5, label=label)
                labeled = True
            if xj <= xmax:
                if labeled:
                    label = None
                ax1.axvline(xj, color='gray', linestyle=':', alpha=0.5, label=label)
    else:
        moment = None

    ax1.legend()

    # Plot the UV spectral curve
    ax2 = fig.add_subplot(gs[0, 1])
    m = xr_curve.get_max_i()
    n = mapping.get_mapped_index(m, xr_curve.x, uv_curve.x)
    uv_jcurve = ssd.uv.get_jcurve(j=n)
    ax2.plot(uv_jcurve.x, uv_jcurve.y, color="C0", label="UV Absorbance at j=%d" % n)
    uv_jslice = trim.uv_slices[0]
    i, j = ij_from_slice(uv_jslice)
    ax2.axvspan(*uv_jcurve.x[[i,j]], color='green', alpha=0.1)
    ax2.axvline(280, color='yellow')  # Assuming 280 nm is the wavelength of interest
    ax2.set_xlabel("Wavelength (nm)")
    ax2.legend()

    # Plot the XR spectral curve
    ax3 = fig.add_subplot(gs[1, 1])
    ax3.set_yscale('log')

    xr_jcurve = ssd.xr.get_jcurve(j=m)
    ax3.plot(xr_jcurve.x, xr_jcurve.y, color="orange", alpha=0.5, label="XR Scattering at j=%d" % m)
    xr_jslice = trim.xr_slices[0]
    i, j = ij_from_slice(xr_jslice)
    ax3.axvspan(*xr_jcurve.x[[i,j]], color='green', alpha=0.1)
    ax3.axvline(0.02, color='yellow')  # Assuming xr_max_x is the Q value of interest
    ax3.set_xlabel(r"Q ($\AA^{-1}$)")
    ax3.legend()

    fig.tight_layout()

    if debug:
        plt.show()

    from molass.PlotUtils.PlotResult import PlotResult
    return PlotResult(fig, (ax1, ax2, ax3),
                      mapping=mapping, xr_curve=xr_curve, uv_curve=uv_curve, mp_curve=mp_curve,
                      moment=moment)


"""
Electron density retrieval tutorial tests with controlled execution order.
Requires: pip install pytest-order
"""

import pytest
from molass.Testing import control_matplotlib_plot

@pytest.mark.order(1)
@control_matplotlib_plot
def test_001_plot_components():
    from molass import requires
    requires('0.6.3')
    from molass_data import SAMPLE1
    from molass.DataObjects import SecSaxsData as SSD
    global decomposition
    ssd = SSD(SAMPLE1)
    trimmed_ssd = ssd.trimmed_copy()
    corrected_ssd = trimmed_ssd.corrected_copy()
    decomposition = corrected_ssd.quick_decomposition(num_components=3)
    decomposition.plot_components() 

output_folder = "temp"

@pytest.mark.order(2)
@control_matplotlib_plot
def test_002_run_denss():
    import warnings
    from molass.SAXS.DenssTools import run_denss
    
    # Suppress deprecation warnings for this specific test
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", DeprecationWarning)
        # Get, for example, the first component's scattering curve as an array
        jcurve_array = decomposition.get_xr_components()[0].get_jcurve_array()
        run_denss(jcurve_array, output_folder=output_folder)

@pytest.mark.order(3)
@control_matplotlib_plot
def test_003_show_mrc():
    import matplotlib.pyplot as plt
    from molass.SAXS.MrcViewer import show_mrc
    show_mrc(output_folder + '/denss_result.mrc');



"""
    PlotUtils.MatrixPlot.py
"""
import numpy as np

def compute_3d_xyz(M, x=None, y=None):
    x_size = M.shape[0]
    n = max(1, x_size//200)
    i = np.arange(0, x_size, n)
    j = np.arange(M.shape[1])
    ii, jj = np.meshgrid(i, j)
    zz = M[ii, jj]
    if x is None:
        x_ = i
    else:
        x_ = x[slice(0, len(x), n)]
    if y is None:
        y = j
    xx, yy = np.meshgrid(x_, y)
    return xx, yy, zz

def simple_plot_3d(ax, M, x=None, y=None, **kwargs):
    """Plot M as a 3D surface on a provided 3D axes.

    The caller is responsible for creating the figure and 3D axes:

        fig, ax = plt.subplots(subplot_kw={'projection': '3d'})
        simple_plot_3d(ax, M, x=q_values, y=frame_numbers)

    Parameters
    ----------
    ax : mpl_toolkits.mplot3d.Axes3D
        An existing 3D axes to plot into.
    M : 2D array-like, shape (n_x, n_y)
        The matrix to render as a surface.
    x : array-like, optional
        Values for the first axis (rows of M). Defaults to integer indices.
    y : array-like, optional
        Values for the second axis (columns of M). Defaults to integer indices.
    **kwargs
        Additional keyword arguments passed to ``ax.plot_surface()``, plus:
        ``view_init`` (dict) — passed to ``ax.view_init()``;
        ``colorbar`` (bool) — add a colorbar if True;
        ``view_arrows`` (bool) — overlay view-direction arrows if True.
    """
    xx, yy, zz = compute_3d_xyz(M, x, y)
    view_init_kwargs = kwargs.pop('view_init', {})
    view_arrows = kwargs.pop('view_arrows', False)
    colorbar = kwargs.pop('colorbar', False)
    sfp = ax.plot_surface(xx, yy, zz, **kwargs)
    if colorbar:
        ax.get_figure().colorbar(sfp, ax=ax)
    if view_arrows:
        from importlib import reload
        import molass.PlotUtils.ViewArrows
        reload(molass.PlotUtils.ViewArrows)
        from molass.PlotUtils.ViewArrows import plot_view_arrows
        plot_view_arrows(ax)
    ax.view_init(**view_init_kwargs)

def contour_plot(ax, M, x=None, y=None, **kwargs):
    xx, yy, zz = compute_3d_xyz(M, x, y)
    ax.contour(xx, yy, zz, **kwargs)


"""
DataUtils.ZipFolder.py

This module is used to zip a folder
"""
import os
import zipfile

def zip_folder(folder, zip_file):
    """
    Zip a folder including the folder name.

    suggested by GitHub Copilot or https://stackoverflow.com/a/1855118

    Parameters
    ----------
    folder : str
        The folder to be zipped.
    zip_file : str
        The output zip file path.
    """
    with zipfile.ZipFile(zip_file, 'w', zipfile.ZIP_DEFLATED) as zipf:
        for root, dirs, files in os.walk(folder):
            for file in files:
                file_path = os.path.join(root, file)
                arcname = os.path.relpath(file_path, os.path.join(folder, '..'))
                zipf.write(file_path, arcname)


"""
Shapes.Sphere.py
"""

import numpy as np

class Sphere:
    """
    Sphere class to represent a spherical density space.

    Attributes
    ----------
    radius : float
        The radius of the sphere.
    center : tuple of float, optional
        The center of the sphere. If None, the center is assumed to be at the
        center of the grid when used in get_condition.
    """
    def __init__(self, radius, center=None):
        """
        Initialize the Sphere object.

        Parameters
        ----------
        radius : float
            The radius of the sphere.
        center : tuple of float, optional
            The center of the sphere. If None, the center is assumed to be at the
            center of the grid when used in get_condition.
        """
        self.radius = radius
        self.center = center

    def get_condition(self, xx, yy, zz, center=None):
        """
        Get the condition for the sphere.

        Parameters
        ----------
        xx : np.ndarray
            The x-coordinates grid.
        yy : np.ndarray
            The y-coordinates grid.
        zz : np.ndarray
            The z-coordinates grid.
        center : tuple of float, optional
            The center of the sphere. If None, the center attribute of the object is used.
            If the center attribute is also None, the center is assumed to be at the center of the grid.
            
        Returns
        -------
        np.ndarray
            A boolean array where points inside the sphere are True.
        """
        if center is None:
            center = self.center
            if center is None:
                assert xx.shape == yy.shape and yy.shape == zz.shape, "xx, yy, zz must have the same shape."
                center = np.array(xx.shape)/2
        return (xx - center[0])**2 + (yy - center[1])**2 + (zz - center[2])**2 <= self.radius**2

    def get_rg(self):
        """
        Calculate the radius of gyration of the sphere.

        Returns
        -------
        float
            The radius of gyration of the sphere.
        """
        return self.radius * np.sqrt(5/3)

    def get_volume(self):
        """
        Calculate the volume of the sphere.

        Returns
        -------
        float
            The volume of the sphere.
        """
        return (4/3) * np.pi * self.radius**3



"""
Baseline.UvdifflBaseline.py
"""
from molass.Baseline.UvBaseline import estimate_uvbaseline_params
from molass.Baseline.LpmBaseline import compute_lpm_baseline

def get_uvdiff_baseline_info(uv_data, pickat=400):
    """Get the parameters and baseline for UVDIFF baseline fitting.

    Note that, in 2D cases, this function is called only once instead of every
    call to the baseline computation.

    Parameters
    ----------
    uv_data : UvData
        The UvData object containing the data for baseline estimation.
    pickat : int, optional
        The index at which to pick the second curve for UVDIFF baseline estimation.
        Default is 400.
        
    Returns
    -------
    params : dict
        A dictionary containing the estimated parameters for the UVDIFF baseline.
    dy : array-like
        The difference between the two curves used for baseline estimation.
    uvdiff_baseline : array-like
        The estimated UVDIFF baseline.
    """
    c1 = uv_data.get_icurve()
    c2 = uv_data.get_icurve(pickat=pickat)
    params, dy, uvdiff_baseline = estimate_uvbaseline_params(c1, c2, pickat=pickat, return_also_baseline=True)
    return params, dy, uvdiff_baseline

def compute_uvdiff_baseline(x, y, uvdiff_info, return_also_params=False, **kwargs):
    """Compute the UVDIFF baseline.
    Parameters
    ----------
    x : array-like
        The x-axis data (wavelengths).

    y : array-like
        The y-axis data (intensities).
    uvdiff_info : tuple
        A tuple containing the parameters, dy, and uvdiff_baseline.
    return_also_params : bool, optional
        If True, the function returns a tuple containing the baseline and a dictionary with parameters.
        If False, it returns only the baseline.
    Returns
    -------
    baseline : array-like
        The computed UVDIFF baseline.
    params : dict, optional
        A dictionary containing the parameters used for baseline computation, if `return_also_params` is True.
    """ 

    # note that uvdiff_info is the return value from get_uvdiff_baseline_info
    params, dy, uvdiff_baseline = uvdiff_info

    lpm_baseline = compute_lpm_baseline(x, y, {})  # Ensure LPM baseline is computed if needed
    ret_baseline = lpm_baseline + uvdiff_baseline
    if return_also_params:
        return ret_baseline, dict(params=params)
    else:
        return ret_baseline


"""
Rigorous.CurrentStateUtils.py
"""
import os
from molass_legacy._MOLASS.SerialSettings import get_setting
from molass_legacy.Optimizer.Scripting import get_params
from molass.LowRank.Decomposition import Decomposition
from molass.SEC.Models.UvComponentCurve import UvComponentCurve
from molass.Mapping.Mapping import Mapping

def construct_decomposition_from_results(run_info, **kwargs):      
    optimizer_folder = get_setting('optimizer_folder')
    wait_for_first_results = kwargs.get('wait_for_first_results', False)
    if wait_for_first_results:
        print(f"Waiting for first results in optimizer folder: {optimizer_folder}")
        import time
        while True:
            jobs_folder = os.path.join(optimizer_folder, "jobs")
            if os.path.exists(jobs_folder):
                jobids = [d for d in os.listdir(jobs_folder) if os.path.isdir(os.path.join(jobs_folder, d))]
                if len(jobids) > 0:
                    job_result_folder = os.path.join(optimizer_folder, "jobs", jobids[-1])
                    result_file = os.path.join(job_result_folder, "callback.txt")
                    if os.path.exists(result_file):
                        try:
                            params = get_params(job_result_folder)
                            break
                        except Exception:
                            pass
            time.sleep(1)
 
    print(f"Loading current decomposition from optimizer folder: {optimizer_folder}")
    jobid = kwargs.get('jobid', None)
    if jobid is None:
        jobs_folder = os.path.join(optimizer_folder, "jobs")
        jobids = [d for d in os.listdir(jobs_folder) if os.path.isdir(os.path.join(jobs_folder, d))]
        jobids.sort()
        jobid = jobids[-1]
    
    job_result_folder = os.path.join(optimizer_folder, "jobs", jobid)
    print(f"Using job id: {jobid}, folder: {job_result_folder}")

    ssd = run_info.ssd
    xr_icurve = ssd.xr.get_icurve()
    if ssd.has_uv():
        uv_icurve = ssd.uv.get_icurve()
    else:
        uv_icurve = None

    params = get_params(job_result_folder)
    optimizer = run_info.optimizer
    separated_params = optimizer.split_params_simple(params)

    # xr_ccurves
    from importlib import reload
    import molass.Rigorous.ComponentUtils
    reload(molass.Rigorous.ComponentUtils)
    from .ComponentUtils import get_xr_ccurves
    xr_ccurves = get_xr_ccurves(optimizer, xr_icurve, separated_params)

    # mapping
    a, b = separated_params[3]
    mapping = Mapping(a, b)

    # uv_ccurves
    uv_params = separated_params[4]
    uv_ccurves = []
    if uv_icurve is None:
        x = xr_icurve.x
    else:
        x = uv_icurve.x
    for xr_ccurve, scale in zip(xr_ccurves, uv_params):
        xr_h = xr_ccurve.get_scale_param()
        uv_ccurves.append(UvComponentCurve(x, mapping, xr_ccurve, scale/xr_h))
    return Decomposition(ssd, xr_icurve, xr_ccurves, uv_icurve, uv_ccurves, **kwargs)


"""
    SEC.Models.GEC.py
"""
import numpy as np
from scipy.special import iv, ive

def gec_monopore_pdf(t, np_, tp_):
    return iv(1, np.sqrt(4*np_*t/tp_)) * np.sqrt(np_/(t*tp_)) * np.exp(-t/tp_-np_)

def robust_gec_monopore_pdf(t, np_, tp_):
    # Bessel functions in Python that work with large exponents
    # https://stackoverflow.com/questions/13726464/bessel-functions-in-python-that-work-with-large-exponents
    #
    # iv(1, np.sqrt(4*np_*t/tp_)) * np.sqrt(np_/(t*tp_)) * np.exp(-t/tp_-np_)
    #
    # ive(v, z) = iv(v, z) * exp(-abs(z.real))
    # iv(v, sq) = ive(v, sq) * exp(sq)

    # val = single_pore_pdf(t, np_, tp_)
    sq = np.sqrt(4*np_*t/tp_)
    val = ive(1, sq) * np.sqrt(np_/(t*tp_)) * np.exp(sq -t/tp_ -np_)
    isnan_val = np.isnan(val)
    val[isnan_val] = 0
    return val

from molass_legacy.SecTheory.SecPDF import FftInvPdf
def gec_monopore_cf(s, np_, tp_):
    # Characteristic function of the GEC monopore model
    return np.exp(np_ * (1/(1 - 1j * tp_ * s) - 1))
 
gec_monopore_numerical_inversion_pdf = FftInvPdf(gec_monopore_cf)


"""
DataUtils.ForwardCompat.py

This module is used to convert old data objects to new ones.
"""
import numpy as np
from scipy.stats import linregress
from scipy.interpolate import UnivariateSpline
from molass.FlowChange.NullFlowChange import CsProxy, NullFlowChange

class CurveProxy:
    """A proxy class for Curve to hold x, y, spline, and peak_info.

    Attributes
    ----------
    x : array-like
        The x-values of the curve.
    y : array-like
        The y-values of the curve.
    spline : UnivariateSpline
        A spline representation of the curve.    
    peak_info : list of lists
        List of peak information, where each peak is represented by a list of indices.
    """
    def __init__(self, x, y, peak_info):
        self.x = x
        self.y = y
        try:
            self.spline = UnivariateSpline(x, y, s=0, ext=3)
        except Exception:
            import logging
            logger = logging.getLogger(__name__)
            logger.warning("Failed to create spline for CurveProxy: len(x)=%d, len(y)=%d", len(x), len(y))
            raise
        self.peak_info = peak_info

class PreRecogProxy:
    """A proxy class for PreRecog to hold flowchange and cs.

    Attributes
    ----------
    flowchange : NullFlowChange
        The flow change object.
    cs : CsProxy
        The calibration slope and intercept proxy.
    """
    def __init__(self, flowchange, cs):
        self.flowchange = flowchange
        self.cs = cs

def get_start_index(slice_):
    """Get the start index from a slice object.

    Parameters
    ----------
    slice_ : slice
        The slice object.

    Returns
    -------
    int
        The start index of the slice.
    """
    j = slice_.start
    if j is None:
        j = 0
    return j

def get_trimmed_curve(curve, slice_, renumber=True, convert_peak_info=True):
    """Get a trimmed version of the curve based on the given slice.

    Parameters
    ----------
    curve : CurveProxy
        The original curve to be trimmed.
    slice_ : slice
        The slice object defining the portion to keep.
    renumber : bool, optional
        If True, renumber the x-values to start from 0, by default True.
    convert_peak_info : bool, optional
        If True, adjust the peak_info indices according to the slice, by default True.

    Returns
    -------
    CurveProxy
        The trimmed curve.
    """
    size = slice_.stop
    if size is None:
        size = len(curve.x)
    j = get_start_index(slice_)
    x, y = curve.get_xy()
    if renumber:
        x_ = np.arange(size - j)
    else:
        x_ = x[slice_]
    y_ = y[slice_]
    if convert_peak_info:
        new_peak_info = []
        for rec in curve.peak_info:
            new_peak_info.append([n - j for n in rec])
    else:
        new_peak_info = None
    return CurveProxy(x_, y_, new_peak_info)

def convert_to_trimmed_prerecog(pre_recog, uv_restrict_list, xr_restrict_list, renumber=True, debug=False):
    """Convert an old PreRecog object to a trimmed PreRecogProxy object.

    Parameters
    ----------
    pre_recog : PreRecog
        The original PreRecog object to be converted.
    uv_restrict_list : list of Restrict
        List of Restrict objects for UV data.
    xr_restrict_list : list of Restrict
        List of Restrict objects for XR data.
    renumber : bool, optional
        If True, renumber the x-values to start from 0, by default True.
    debug : bool, optional
        If True, enable debug mode, by default False.
        
    Returns
    -------
    PreRecogProxy
        The converted and trimmed PreRecogProxy object.
    """
    
    if debug:
        print("convert_to_trimmed_prerecog")
        print("uv_restrict_list=", uv_restrict_list)
        print("xr_restrict_list=", xr_restrict_list)
    
    fc = pre_recog.flowchange

    def get_slice_from_restrict_list(restrict_list):
        if restrict_list is None or len(restrict_list) == 0 or restrict_list[0] is None:
            return slice(None)
        else:
            return restrict_list[0].get_slice()

    uv_slice = get_slice_from_restrict_list(uv_restrict_list)
    trimmed_uv_curves = []
    for k, curve in enumerate([fc.a_curve, fc.a_curve2]):
        trimmed_uv_curves.append(get_trimmed_curve(curve, uv_slice, renumber=renumber, convert_peak_info=k == 0))

    xr_slice = get_slice_from_restrict_list(xr_restrict_list)
    old_cs = pre_recog.cs
    trimmed_xr_curve = get_trimmed_curve(old_cs.x_curve, xr_slice, renumber=renumber)

    xr_x = old_cs.x_curve.x
    uv_x = fc.a_curve.x
    X = xr_x[[0,-1]]
    slope = old_cs.slope
    intercept = old_cs.intercept
    Y = slope * X + intercept
    i = get_start_index(xr_slice)
    j = get_start_index(uv_slice)
    X_ = X - xr_x[i]
    Y_ = Y - uv_x[j]
    slope_, intercept_ = linregress(X_, Y_)[0:2]
    new_cs = CsProxy(slope_, intercept_)

    return PreRecogProxy(NullFlowChange(*trimmed_uv_curves, trimmed_xr_curve), new_cs)



"""
    Mapping.PeakMatcher.py
"""
import numpy as np
from itertools import combinations
from scipy.stats import linregress

def combination_pairs(m, n):
    assert m != n
    more_indeces = np.arange(max(m, n))
    num = min(m, n)
    less_indeces = np.arange(num)
    for c in combinations(more_indeces, num):
        if m > n:
            yield list(c), less_indeces
        else:
            yield less_indeces, list(c)

def evaluate_matching(xr_x, uv_x, xr_peaks, uv_peaks):
    x = xr_x[xr_peaks]
    y = uv_x[uv_peaks]
    slope, intercept, r_value, p_value, std_err = linregress(x, y)

    mapped_xr_ends = []
    for px in xr_x[[0, -1]]:
        mapped_xr_ends.append(px*slope + intercept)
    uv_minx = max(uv_x[0], mapped_xr_ends[0])
    uv_maxx = min(uv_x[-1], mapped_xr_ends[1])
    # y = ax + b
    # x = (y - b)/a
    a_ = 1/slope
    b_ = -intercept/slope
    mapped_uv_ends = []
    for px in uv_minx, uv_maxx:
        mapped_uv_ends.append(px*a_ + b_)
    xr_minx = max(xr_x[0], mapped_uv_ends[0])
    xr_maxx = max(xr_x[-1], mapped_uv_ends[1])
    covered_ratio = (xr_maxx - xr_minx)/(xr_x[-1] - xr_x[0])
    score = 1/covered_ratio

    if len(x) > 2:
        score *= p_value
    else:
        # do not use p_value since it is zero
        pass

    return score

def select_matching_peaks(xr_curve, xr_peaks, uv_curve, uv_peaks, debug=False):
    """
    Select matching peaks between XR and UV curves.

    For the evaluation using the weights, see the debugging info by using BSA_DATA.

    Parameters
    ----------
    xr_curve : Curve
        The XR curve object.
    xr_peaks : array-like
        The indices of the peaks in the XR curve.
    uv_curve : Curve
        The UV curve object.
    uv_peaks : array-like
        The indices of the peaks in the UV curve.
    debug : bool, optional
        If True, print debug information.

    Returns
    -------
    tuple
        A tuple containing the selected matching peaks for XR and UV curves.
    """
    xr_x = xr_curve.x
    xr_y = xr_curve.y
    uv_x = uv_curve.x
    uv_y = uv_curve.y
    if debug:
        print("len(xr_peaks)=", len(xr_peaks))
        print("len(uv_peaks)=", len(uv_peaks))
    xr_peaks = np.asarray(xr_peaks)
    uv_peaks = np.asarray(uv_peaks)
    xr_weights = xr_y[xr_peaks]
    xr_weights = xr_weights / np.sum(xr_weights)
    uv_weights = uv_y[uv_peaks]
    uv_weights = uv_weights / np.sum(uv_weights)

    # evaluate all the combination pairs
    score_recs = []
    for index1, index2 in combination_pairs(len(xr_peaks), len(uv_peaks)):
        score = evaluate_matching(xr_x, uv_x, xr_peaks[index1], uv_peaks[index2])
        score *= 1/(np.sum(xr_weights[index1]) * np.sum(uv_weights[index2]))
        if debug:
            print(index1, index2, score)
        score_recs.append((index1, index2, score))
    
    score_recs = sorted(score_recs, key=lambda x: x[2])
    index1, index2, score = score_recs[0]   # the record with smallest p_value
    return xr_peaks[index1], uv_peaks[index2]


"""
    test Trimming
"""
from molass import get_version
get_version(toml_only=True)     # to ensure that the current repository is used
from molass_data import SAMPLE2
from molass.Testing import control_matplotlib_plot, is_interactive

@control_matplotlib_plot
def test_010_PKS():
    from molass.DataObjects import SecSaxsData as SSD
    ssd = SSD(SAMPLE2)
    ssd.plot_trimming(debug=is_interactive())
    trimmed_ssd = ssd.trimmed_copy()
    trimmed_ssd.plot_trimming(debug=is_interactive())

if __name__ == "__main__":
    test_010_PKS()


"""
    SEC.Models.EdmComponentCurve.py

"""
import numpy as np
from molass_legacy.Models.RateTheory.EDM import edm_impl
from molass.LowRank.ComponentCurve import ComponentCurve

class EdmColumn:
    """
    A class to represent an EDM column.

    Attributes
    ----------
    params : tuple
        The parameters of the EDM column (N, T, me, mp, x0, tI, N0, poresize, timescale).
    """
    def __init__(self, params):
        """
        Initializes the EDM column.

        Parameters
        ----------
        params : tuple
            The column parameters (N, T, me, mp, x0, tI, N0, poresize, timescale).
        """
        self.params = params

    def get_params(self):
        """
        Returns the parameters of the EDM column.

        Returns
        -------
        tuple
            The parameters of the EDM column.
        """
        return self.params

class EdmComponentCurve(ComponentCurve):
    """
    A class to represent an EDM component curve.

    Attributes
    ----------
    x : array-like
        The x values.
    params : tuple
        The parameters of the EDM column (N, T, me, mp, x0,
    """
    def __init__(self, x, params):
        """
        Initializes the EDM component curve.
        Parameters
        ----------
        x : array-like
            The x values.
        params : tuple
            The column parameters (N, T, me, mp, x0, tI, N0, poresize, timescale).
        """
        self.x = x
        self.params = params
        self.moment = None
        self.model = 'edm'
    
    def get_y(self, x=None):
        """
        Returns the y values for the given x values.

        Parameters
        ----------
        x : array-like or None, optional
            The x values to get the y values for. If None, uses the object's x values.

        Returns
        -------
        array-like

        """
        if x is None:
            x = self.x
        return edm_impl(x, *self.params)
    
    def get_peak_top_x(self):
        """
        Returns the x value at the peak top.

        Raises
        ------
        NotImplementedError
            If the peak top x calculation is not implemented for the current model.
        """
        raise NotImplementedError("Peak top x calculation is not implemented for SDM model.")


"""
    PlotUtils.ViewArrows.py
"""

def plot_view_arrows(ax, use_quiver=False, debug=False):
    """
    Plot view arrows on a 3D axis.
    This function adds arrows to indicate the elutional and spectral views.
    """
    if debug:
        from importlib import reload
        import molass.PlotUtils.Arrow3D
        reload(molass.PlotUtils.Arrow3D)
    from molass.PlotUtils.Arrow3D import add_arrow3D

    xmin, xmax = ax.get_xlim()
    ymin, ymax = ax.get_ylim()
    zmin, zmax = ax.get_zlim()
    ax.set_xlim(xmin, xmax)
    ax.set_ylim(ymin, ymax)
    ax.set_zlim(zmin, zmax)
    if debug:
        print(f"X: {xmin}, {xmax}")
        print(f"Y: {ymin}, {ymax}")
        print(f"Z: {zmin}, {zmax}")

    def get_coordinates(x, y, z):
        return (x * (xmax - xmin) + xmin,
                y * (ymax - ymin) + ymin,
                z * (zmax - zmin) + zmin)

    def get_direction(x, y, z):
        return (x * (xmax - xmin),
                y * (ymax - ymin),
                z * (zmax - zmin))

    xz_view_point = get_coordinates(0.5, -0.6, 0)
    xz_head_vector = get_direction(0, 0.4, 0)
    yz_view_point = get_coordinates(1.6, 0.5, 0)
    yz_head_vector = get_direction(-0.3, 0, 0)
    
    if use_quiver:
        # Use quiver for 3D arrows
        # Note: currently not available possibly due to a bug.
        xz_arrow_length = 0.3*(xmax - xmin)
        yz_arrow_length = 0.3*(ymax - ymin)
        ax.quiver(*xz_view_point, *xz_head_vector, color='orange', length=xz_arrow_length, arrow_length_ratio=0.2, normalize=True)
        ax.quiver(*yz_view_point, *yz_head_vector, color='orange', length=yz_arrow_length, arrow_length_ratio=0.2, normalize=True)
    else:
        add_arrow3D()
        arrowstyle="-|>"
        arrow_spec = dict(color='orange', arrowstyle=arrowstyle, linewidth=3, linestyle='-', mutation_scale=20)
        ax.arrow3D(*xz_view_point, *xz_head_vector, **arrow_spec)
        ax.arrow3D(*yz_view_point, *yz_head_vector, **arrow_spec)

    ax.text(*xz_view_point, "Elutional View", color='orange', va='top', ha='center', fontsize=14)
    ax.text(*yz_view_point, "Spectral View", color='orange', va='top', ha='center', fontsize=14)


"""
    DataUtils.AnomalyHandlers.py
"""
import numpy as np
from scipy.signal import find_peaks

BUBBLE_WIDTH_RANGE = (1, 5)
BUBBLE_SEARCH_WIDTH = 10
BUBBLE_GRADIENT_LIMIT = 0.03

def bubble_check_impl(y, debug=False):
    """
    Check for bubbles in the intensity array.
    
    Parameters
    ----------
    y : np.ndarray
        The intensity array to be checked for bubbles.       
    debug : bool
        If True, debug information will be printed.

    Returns
    -------
    bubbles : np.ndarray
        The indices of the detected bubbles.
    """
    max_y = np.max(y)
    height = max_y*0.9
    width = BUBBLE_WIDTH_RANGE
    prominence = height
    peaks, _ = find_peaks(y, height=height, width=width, prominence=prominence)
    if len(peaks) > 0:
        assert len(peaks) == 1
        m = peaks[0]
        start = max(0, m - BUBBLE_SEARCH_WIDTH)
        stop = min(m + BUBBLE_SEARCH_WIDTH, len(y))
        y_ = y[start:stop]
        gy = np.abs(np.gradient(y_))/max_y
        bubbles = start + np.where(gy > BUBBLE_GRADIENT_LIMIT)[0]
    else:
        bubbles = np.array([], dtype=int)
    if debug:
        import matplotlib.pyplot as plt
        x = np.arange(len(y))
        fig, ax = plt.subplots()
        ax.set_title("find_peaks(y, height=%.3g, width=%s, prominence=%.3g)" % (height, width, prominence))
        ax.plot(x, y)
        ax.plot(x[bubbles], y[bubbles], 'o', color='orange', alpha=0.5)
        ax.plot(x[peaks], y[peaks], 'o', color='red', alpha=0.5)
        axt = ax.twinx()
        if False:
            x_ = x[start:stop]
            axt.plot(x_, gy, 'x')
    return bubbles

def remove_bubbles_impl(intensity_array, to_be_excluded, excluded_set, debug=False):
    """
    Exclude bubbles from the intensity array.

    Parameters
    ----------
    intensity_array : np.ndarray
        The intensity array to be modified.
    to_be_excluded : np.ndarray
        The indices of the bubbles to be excluded.
    excluded_set : set
        The set of excluded indices.
    debug : bool
        If True, debug information will be printed.

    Returns
    -------
    excluded_set : set
        The set of excluded indices.

    """
    from_ = None
    for i in to_be_excluded:
        if from_ is None:
            from_ = i
        else:
            if i > last + 1:
                exclude_bubble_impl(intensity_array, from_, last, excluded_set)
                from_ = i

        last = i
    if from_ is not None:
        exclude_bubble_impl(intensity_array, from_, last, excluded_set)

    return excluded_set

def exclude_bubble_impl(intensity_array, from_, to_, excluded_set):
    """
    Exclude a bubble from the intensity array.
    """
    print( 'exclude_bubble_impl: ', from_, to_ )
    size = intensity_array.shape[0]
    if from_ == 0:
        j = to_ + 1
        for i in range( from_, j ):
            intensity_array[i, :, 1:] = intensity_array[j, :, 1:]
            excluded_set.add(i)
    elif to_ == size - 1:
        j = from_ - 1
        for i in range( from_, size ):
            intensity_array[i, :, 1:] = intensity_array[j, :, 1:]
            excluded_set.add(i)
    else:
        lower = from_ - 1
        upper = to_ + 1
        lower_intensity = intensity_array[lower, :, : ]
        upper_intensity = intensity_array[upper, :, : ]
        width = upper - lower
        for i in range( 1, width ):
            w = i/width
            intensity = ( 1 - w  ) * lower_intensity[ :, 1: ] + w * upper_intensity[ :, 1: ]
            intensity_array[lower+i, :, 1:] = intensity
            excluded_set.add(lower+i)


"""
Rigorous.ComponentUtils.py
"""

def get_egh_xr_ccurves(optimizer, xr_icurve, separated_params):
    from molass.LowRank.ComponentCurve import ComponentCurve
    xr_params = separated_params[0]
    x = xr_icurve.x
    xr_ccurves = []
    for p in xr_params:
        xr_ccurves.append(ComponentCurve(x, p))
    return xr_ccurves

def get_sdm_xr_ccurves(optimizer, xr_icurve, separated_params):
    from molass_legacy.Models.Stochastic.DispersivePdf import DEFUALT_TIMESCALE
    from molass.SEC.Models.SdmComponentCurve import SdmColumn, SdmComponentCurve
    xr_params = separated_params[0]
    rg_params = separated_params[2]
    N, K, x0, poresize, N0, tI = separated_params[-1]
    T = K/N
    me = mp = 1.5
    column = SdmColumn([N, T, me, mp, x0, tI, N0, poresize, DEFUALT_TIMESCALE])
    x = xr_icurve.x
    xr_ccurves = []
    for scale, rg in zip(xr_params, rg_params):
        xr_ccurves.append(SdmComponentCurve(x, column, rg, scale))
    return xr_ccurves

def get_edm_xr_ccurves(optimizer, xr_icurve, separated_params):
    raise NotImplementedError("EDM XR component curve extraction not implemented yet.")

def get_xr_ccurves(optimizer, xr_icurve, separated_params):
    model_name = optimizer.get_model_name()
    if model_name == 'EGH':
        return get_egh_xr_ccurves(optimizer, xr_icurve, separated_params)
    elif model_name == 'SDM':
        return get_sdm_xr_ccurves(optimizer, xr_icurve, separated_params)
    elif model_name == 'EDM':
        return get_edm_xr_ccurves(optimizer, xr_icurve, separated_params)
    else:
        raise ValueError(f"Unknown model_name: {model_name}")


"""
    SurveyUtils.UvPickvalue.py
"""

def get_pickvalue(in_folder, wvector):
    """Determine the UV pickvalue based on the input folder name and wvector.
    
    Parameters
    ----------
    in_folder : str
        The input folder name.
    wvector : array-like
        The wavelength vector.

    Returns
    -------
    pickvalue : float
        The determined pickvalue.
    """
    if in_folder.find('OAGIwyatt_01') >= 0 or in_folder.find('OA_Ald_Fer'):
        pickvalue = 400
    elif in_folder.find('20190315_1') >= 0:
        pickvalue = 550
    else:
        pickvalue = wvector[-5]
    return pickvalue


"""
    LowRank.PairedRange.py

    This module contains the class PairedRange, which is used to store information
    about the valid elution ranges for the analysis report.
"""

class PairedRange:
    """A class to represent a paired range.
    It contains a pair of ranges which correspond to the ascending and descending parts of the peak.

    Attributes
    ----------
    ranges : list of tuples
        A list of tuples representing the ranges. Each tuple contains two integers (start, end).
    peak_index : int or None
        The index of the peak. It can be None if not specified.
    elm_recs : list or None
        The elution records associated with the ranges. It can be None if not specified.
    """
    def __init__(self, range_, minor=False, peak_index=None, elm_recs=None):
        self.peak_index = peak_index
        if minor:
            ranges = [range_]
        else:
            if peak_index is None:
                peak_index = (range_[0] + range_[1])//2
            ranges = [(range_[0], peak_index), (peak_index, range_[1])]

        self.ranges = ranges
        self.elm_recs = elm_recs

    def get_fromto_list(self):
        """
        for backward compatibility to molass_legacy.AnalysisRangeInfo.PaoiredRange,
        returns a list of tuples
        """
        return self.ranges

    def as_list(self, k):
        """Returns a list of PeakInfo and RangeInfo objects."""
        from molass_legacy.DataStructure.PeakInfo import PeakInfo
        peakinfo = PeakInfo(k,
                         self.peak_index,
                         self.elm_recs
                         )  
        return [peakinfo] + self.ranges

    def is_minor(self):
        """Returns True if the PairedRange is a minor peak (i.e., has only one range)."""
        return len(self.ranges) == 1

    def __len__(self):
        return len(self.ranges)

    def __iter__(self):
        for range_ in self.ranges:
            yield range_

    def __str__(self):
        return str(self.ranges)

    def __repr__(self):
        return str(self.ranges)

def convert_to_list_pairedranges(pairedranges):
    ret_list = []
    for prange in pairedranges:
        for range_ in prange.ranges:
            ret_list += [range_]
    return ret_list


"""
    test Custom Data Loading and Visualization
"""
import os
import matplotlib.pyplot as plt
from molass import get_version
get_version(toml_only=True)     # to ensure that the current repository is used
from molass.Local import get_local_settings
local_settings = get_local_settings()
TUTORIAL_DATA = local_settings['TUTORIAL_DATA']

def test_010_load_uv():
    from molass.DataUtils.UvLoader import load_uv
    file_path = os.path.join(TUTORIAL_DATA)
    uvM, wvector = load_uv(file_path)
    assert uvM.shape == (318, 574)
    assert wvector.shape == (318,)

def test_020_load_xr():
    from molass.DataUtils.XrLoader import load_xr
    xr_array, datafiles = load_xr(TUTORIAL_DATA)
    assert xr_array.shape == (287, 1176, 3)

def test_030_plot():
    from molass.DataObjects import SecSaxsData as SSD
    ssd = SSD(TUTORIAL_DATA)
    ssd.plot_compact(debug=is_interactive())


"""
    test SSD
"""
from molass import get_version
get_version(toml_only=True)     # to ensure that the current repository is used
from molass_data import SAMPLE1

import pytest
from molass.DataObjects import SecSaxsData as SSD

ssd_instance = SSD(SAMPLE1)

def test_01_constructor():
    assert ssd_instance is not None, "SSD object should not be None"
    assert hasattr(ssd_instance, 'xr'), "SSD object should have 'xr' attribute"
    assert hasattr(ssd_instance, 'uv'), "SSD object should have 'uv' attribute"

def test_02_uv_friendly_aliases():
    uv = ssd_instance.uv
    # wavelengths and frames are human-readable aliases
    import numpy as np
    assert np.array_equal(uv.wavelengths, uv.iv), "wavelengths should equal iv"
    assert np.array_equal(uv.frames, uv.jv), "frames should equal jv"
    assert np.array_equal(uv.wavelengths, uv.wv), "wavelengths should equal wv"
    # M shape should match
    assert uv.M.shape == (len(uv.wavelengths), len(uv.frames)), \
        "M shape should be (wavelengths, frames)"

def test_04_ssd_data_alias():
    xr = ssd_instance.xr
    uv = ssd_instance.uv
    import numpy as np
    # .data should be identical to .M for both XrData and UvData
    assert np.array_equal(xr.data, xr.M), "xr.data should equal xr.M"
    assert np.array_equal(uv.data, uv.M), "uv.data should equal uv.M"

def test_05_repr():
    xr = ssd_instance.xr
    uv = ssd_instance.uv
    xr_repr = repr(xr)
    uv_repr = repr(uv)
    # repr should contain class name and shape info
    assert "iv=" in xr_repr and "jv=" in xr_repr, f"xr repr missing shape info: {xr_repr}"
    assert "wavelengths=" in uv_repr and "frames=" in uv_repr, f"uv repr missing shape info: {uv_repr}"
    assert "nm" in uv_repr, f"uv repr should include wavelength range in nm: {uv_repr}"

def test_06_wavelength_range():
    uv = ssd_instance.uv
    wl_min, wl_max = uv.wavelength_range
    assert wl_min < wl_max, "wavelength_range min should be less than max"
    assert wl_min == uv.wavelengths.min(), "wavelength_range min should match wavelengths.min()"
    assert wl_max == uv.wavelengths.max(), "wavelength_range max should match wavelengths.max()"

def test_03_plot_3d():
    plot_result = ssd_instance.plot_3d()
    assert plot_result is not None, "Plot result should not be None"
    plot_result = ssd_instance.plot_3d(uv_only=True)
    assert plot_result is not None, "Plot result should not be None"
    plot_result = ssd_instance.plot_3d(xr_only=True)
    assert plot_result is not None, "Plot result should not be None"

def test_07_rgcurve_y_is_float_nan():
    """RgCurve.y must be a float array with nan for failed fits — issue #22."""
    import numpy as np
    trimmed   = ssd_instance.trimmed_copy()
    corrected = trimmed.corrected_copy()
    rgcurve   = corrected.xr.compute_rgcurve()
    # y must be a float array, not an object array
    assert rgcurve.y.dtype == float, \
        f"RgCurve.y should be float dtype, got {rgcurve.y.dtype}"
    # failed fits must be nan, not None
    assert None not in rgcurve.y, \
        "RgCurve.y should not contain None (use nan instead)"
    # at least some frames should have valid Rg
    assert np.any(np.isfinite(rgcurve.y) & (rgcurve.y > 0)), \
        "RgCurve.y should contain at least some valid positive Rg values"


def test_08_get_baseline2d_no_stdout(capsys):
    """get_baseline2d() must not print diagnostic messages to stdout — issue #23."""
    trimmed = ssd_instance.trimmed_copy()
    trimmed.xr.get_baseline2d()
    captured = capsys.readouterr()
    assert captured.out == "", \
        f"get_baseline2d() should not print to stdout, got: {captured.out!r}"


def test_09_buffit_baseline():
    """'buffit' method returns a baseline with positive_ratio lower than 'linear' — issue #24."""
    import numpy as np

    def _positive_ratio_mean(M):
        """Mean fraction of positive intensities across all frames, simple unweighted."""
        return np.mean(M > 0)

    trimmed = ssd_instance.trimmed_copy()

    # linear baseline (adaptive p_final)
    trimmed.xr.baseline_method = 'linear'
    b_linear = trimmed.xr.get_baseline2d()
    ratio_linear = _positive_ratio_mean(trimmed.xr.M - b_linear)

    # buffit baseline
    trimmed.xr.baseline_method = 'buffit'
    b_buffit = trimmed.xr.get_baseline2d()
    ratio_buffit = _positive_ratio_mean(trimmed.xr.M - b_buffit)

    assert b_buffit.shape == trimmed.xr.M.shape, \
        "buffit baseline shape should match M"
    assert ratio_buffit < ratio_linear, \
        f"buffit positive_ratio ({ratio_buffit:.3f}) should be lower than linear ({ratio_linear:.3f})"


def test_10_buffit_otsu():
    """Otsu adaptive threshold (default) is at least as good as fixed threshold=0.10 — issue #25."""
    import numpy as np
    from molass.Baseline.BuffitBaseline import BUFFIT_THRESHOLD

    def _positive_ratio_mean(M):
        return np.mean(M > 0)

    trimmed = ssd_instance.trimmed_copy()
    trimmed.xr.baseline_method = 'buffit'

    # Otsu (default: threshold=None)
    b_otsu = trimmed.xr.get_baseline2d()
    ratio_otsu = _positive_ratio_mean(trimmed.xr.M - b_otsu)

    # Fixed threshold=0.10
    b_fixed = trimmed.xr.get_baseline2d(threshold=BUFFIT_THRESHOLD)
    ratio_fixed = _positive_ratio_mean(trimmed.xr.M - b_fixed)

    assert b_otsu.shape == trimmed.xr.M.shape, \
        "Otsu buffit baseline shape should match M"
    assert ratio_otsu <= ratio_fixed + 0.01, \
        f"Otsu positive_ratio ({ratio_otsu:.3f}) should be no worse than fixed ({ratio_fixed:.3f})"


def test_11_q_values_frame_indices_aliases():
    """SsMatrixData.q_values and .frame_indices are aliases for iv and jv — issue #28."""
    import numpy as np
    xr = ssd_instance.xr
    # read aliases
    assert np.array_equal(xr.q_values, xr.iv), "q_values should equal iv"
    assert np.array_equal(xr.frame_indices, xr.jv), "frame_indices should equal jv"
    # setter: modify via alias, check original
    orig_iv = xr.iv.copy()
    xr.q_values = orig_iv * 2
    assert np.array_equal(xr.iv, orig_iv * 2), "setting q_values should update iv"
    xr.q_values = orig_iv  # restore


def test_12_get_bpo_ideal():
    """SsMatrixData.get_bpo_ideal() returns dataset-relative ideal positive_ratio — issue #29."""
    xr = ssd_instance.xr
    bpo_ideal = xr.get_bpo_ideal()
    assert isinstance(bpo_ideal, float), f"bpo_ideal should be float, got {type(bpo_ideal)}"
    assert 0 < bpo_ideal < 1, f"bpo_ideal should be in (0, 1), got {bpo_ideal}"
    # SAMPLE1 is a well-behaved dataset; bpo_ideal should be well above 0.5
    assert bpo_ideal > 0.5, f"SAMPLE1 bpo_ideal should be > 0.5, got {bpo_ideal}"


def test_13_lpm_baseline_mask():
    """compute_lpm_baseline supports mask parameter for frame exclusion — issue #30."""
    import numpy as np
    from molass.Baseline.LpmBaseline import compute_lpm_baseline
    xr = ssd_instance.xr
    x = xr.jv.astype(float)
    y = xr.M[0, :]  # first q-row

    # Without mask
    bl_full, params_full = compute_lpm_baseline(x, y, return_also_params=True)
    assert bl_full.shape == x.shape, "baseline should have same shape as x"

    # With mask: exclude middle 20% of frames
    n = len(x)
    mask = np.ones(n, dtype=bool)
    mask[n//3 : 2*n//3] = False
    bl_masked, params_masked = compute_lpm_baseline(x, y, return_also_params=True, mask=mask)

    # Masked baseline still covers ALL frames
    assert bl_masked.shape == x.shape, "masked baseline should have same shape as x"
    # Slopes should differ (different fitting data)
    assert params_full['slope'] != params_masked['slope'], \
        "masked and unmasked slopes should differ when peak region is excluded"


def test_14_get_snr_weights():
    """get_snr_weights() returns per-q-row SNR array — issue #38."""
    import numpy as np
    xr = ssd_instance.trimmed_copy().xr
    w = xr.get_snr_weights()
    assert w.shape == (len(xr.iv),), f"shape should be (n_q,), got {w.shape}"
    assert (w >= 0).all(), "weights must be non-negative"
    assert w.sum() > 0, "weights should not all be zero"
    # Low-q rows should generally have higher SNR than high-q rows
    n_q = len(w)
    assert w[:n_q // 4].mean() > w[-n_q // 4:].mean(), \
        "low-q SNR should exceed high-q SNR on average"


def test_15_get_positive_ratio():
    """get_positive_ratio() returns SNR-weighted positive_ratio — issue #38."""
    import numpy as np
    xr = ssd_instance.trimmed_copy().xr
    bl = xr.get_baseline2d()

    pr_snr = xr.get_positive_ratio(bl)                     # default = 'snr'
    pr_uni = xr.get_positive_ratio(bl, weighting='uniform')
    assert 0 < pr_snr < 1, f"positive_ratio (snr) should be in (0,1), got {pr_snr}"
    assert 0 < pr_uni < 1, f"positive_ratio (uni) should be in (0,1), got {pr_uni}"
    # SNR and uniform should generally differ
    assert pr_snr != pr_uni, "SNR-weighted and uniform should produce different values"


def test_16_get_bpo_ideal_snr():
    """get_bpo_ideal() defaults to SNR-weighted, and accepts 'uniform' — issue #38."""
    xr = ssd_instance.trimmed_copy().xr
    bpo_snr = xr.get_bpo_ideal()                           # default = 'snr'
    bpo_uni = xr.get_bpo_ideal(weighting='uniform')
    assert 0 < bpo_snr < 1, f"bpo_ideal (snr) should be in (0,1), got {bpo_snr}"
    assert 0 < bpo_uni < 1, f"bpo_ideal (uni) should be in (0,1), got {bpo_uni}"
    assert bpo_snr > 0.5, f"SAMPLE1 bpo_ideal (snr) should be > 0.5, got {bpo_snr}"
    # SNR-weighted bpo_ideal is typically NOT the same as uniform
    assert bpo_snr != bpo_uni, "SNR and uniform bpo_ideal should differ"


def test_17_get_ideal_positive_ratio_alias():
    """get_ideal_positive_ratio() is an alias for get_bpo_ideal() -- issue #40."""
    xr = ssd_instance.trimmed_copy().xr
    assert xr.get_ideal_positive_ratio() == xr.get_bpo_ideal()
    assert xr.get_ideal_positive_ratio(weighting='uniform') == xr.get_bpo_ideal(weighting='uniform')


def test_18_evaluate_baseline():
    """evaluate_baseline() returns namedtuple with positive_ratio, ideal, delta -- issue #39."""
    xr = ssd_instance.trimmed_copy().xr
    bl = xr.get_baseline2d()
    result = xr.evaluate_baseline(bl)
    assert hasattr(result, 'positive_ratio')
    assert hasattr(result, 'ideal')
    assert hasattr(result, 'delta')
    assert result.delta == abs(result.positive_ratio - result.ideal)
    assert 0 < result.positive_ratio < 1
    assert 0 < result.ideal < 1
    # Should match individual calls
    assert result.positive_ratio == xr.get_positive_ratio(bl)
    assert result.ideal == xr.get_bpo_ideal()


"""
SEC.Models.SDM.py
"""
class SDM:
    """
    Stochastic Dispersive Model (SDM) for SEC data analysis.
    """
    def __init__(self, **kwargs):
        """ Initialize the SDM model with given parameters.
        Parameters
        ----------
        kwargs : dict
            Additional parameters for the SDM model.
        """
        self.kwargs = kwargs

    def optimize_decomposition(self, decomposition, **kwargs):
        """ Optimize the given decomposition using the SDM model.

        Parameters
        ----------
        decomposition : Decomposition
            The initial decomposition to be optimized.
        kwargs : dict
            Additional parameters for the optimization process.

        Returns
        -------
        Decomposition
            The optimized decomposition.
        """
        debug = kwargs.get('debug', False)
        if debug:
            from importlib import reload
            import molass.SEC.Models.SdmEstimator
            reload(molass.SEC.Models.SdmEstimator)
            import molass.SEC.Models.SdmOptimizer
            reload(molass.SEC.Models.SdmOptimizer)
            import molass.SEC.Models.UvOptimizer
            reload(molass.SEC.Models.UvOptimizer)
        from molass.SEC.Models.SdmEstimator import estimate_sdm_column_params
        from molass.SEC.Models.SdmOptimizer import optimize_sdm_xr_decomposition, adjust_rg_and_poresize
        from molass.SEC.Models.UvOptimizer import optimize_uv_decomposition

        env_params = estimate_sdm_column_params(decomposition, **kwargs)
        model_params = kwargs.pop('model_params', None)
        new_xr_ccurves = optimize_sdm_xr_decomposition(decomposition, env_params, model_params=model_params, **kwargs)
        if decomposition.uv is None:
            from molass.Decompose.XrOnlyUtils import make_dummy_uv_ccurves
            new_uv_ccurves = make_dummy_uv_ccurves(decomposition.ssd, new_xr_ccurves)
        else:
            new_uv_ccurves = optimize_uv_decomposition(decomposition, new_xr_ccurves, **kwargs)
        sdm_decomposition = decomposition.copy_with_new_components(new_xr_ccurves, new_uv_ccurves)

        rgcurve = kwargs.pop('rgcurve', None)
        adjust_rg_and_poresize(sdm_decomposition, rgcurve=rgcurve)

        if debug:
            import matplotlib.pyplot as plt
            from molass.PlotUtils.DecompositionPlot import plot_elution_curve            
            fig, axes = plt.subplots(nrows=2, ncols=2, figsize=(12, 10))
            fig.suptitle('Optimization Debug Plots')
            if decomposition.uv is not None:
                plot_elution_curve(axes[0, 0], decomposition.uv_icurve, decomposition.uv_ccurves, title="EGH Elution Curves for UV", ylabel="Absorbance")
                plot_elution_curve(axes[1, 0], sdm_decomposition.uv_icurve, sdm_decomposition.uv_ccurves, title="SDM Elution Curves for UV", ylabel="Absorbance")
            plot_elution_curve(axes[0, 1], decomposition.xr_icurve, decomposition.xr_ccurves, title="EGH Elution Curves for XR", ylabel="Scattering Intensity")    
            plot_elution_curve(axes[1, 1], sdm_decomposition.xr_icurve, sdm_decomposition.xr_ccurves, title="SDM Elution Curves for XR", ylabel="Scattering Intensity")
            fig.tight_layout()
        return sdm_decomposition


"""
Reports.ReportRange.py
"""
import numpy as np

MINOR_COMPONENT_MAX_PROP = 0.2

def make_v1report_ranges_impl(decomposition, ssd, mapped_curve, area_ratio, concentration_datatype, debug=False):
    """
    Make V1 report ranges from the decomposition and mapped curve.

    molass_legacy scheme:

    molass library scheme:
        decomp_result = Backward.DecompResultAdapter.adapted_decomp_result(...)

    Parameters
    ----------
    decomposition : Decomposition
        The decomposition object containing the components.
    ssd : SecSaxsData
        The SecSaxsData object containing the data.
    mapped_curve : ICurve
        The mapped curve used for decomposition.
    area_ratio : float
        The area ratio used for determining ranges.
    concentration_datatype : int
        The concentration data type (1 for X-ray measured data, else for UV measured data).
    debug : bool, optional
        If True, print debug information.

    Returns
    -------
    list of PairedRange
        The list of paired ranges for the components.
    """
    if debug:
        from importlib import reload
        import molass.Backward.DecompResultAdapter
        reload(molass.Backward.DecompResultAdapter)
        import molass_legacy.Decomposer.DecompUtils
        reload(molass_legacy.Decomposer.DecompUtils)
    from molass.Backward.DecompResultAdapter import adapted_decomp_result
    from molass_legacy.Decomposer.DecompUtils import make_range_info_impl 
    # task: concentration_datatype must have been be set before calling this function.

    decomp_result= adapted_decomp_result(decomposition, ssd, mapped_curve, debug=debug)
    if concentration_datatype == 1:     # Xray measured data
        opt_recs_ = decomp_result.opt_recs   # will not be used, currenctly
    else:
        opt_recs_ = decomp_result.opt_recs_uv

    control_info = decomp_result.get_range_edit_info()
    # print("editor_ranges=", control_info.editor_ranges)
    # print("select_matrix=", control_info.select_matrix)
    # print("top_x_list=", control_info.top_x_list)

    legacy_ranges = make_range_info_impl(opt_recs_, control_info)
    # print("legacy_ranges=", legacy_ranges)

    elm_recs = decomp_result.opt_recs
    elm_recs_uv = decomp_result.opt_recs_uv

    components = decomposition.get_xr_components()
    # components = decomposition.get_uv_components()

    if debug:
        import matplotlib.pyplot as plt
        fig, ax = plt.subplots()
        for comp in components:
            icurve = comp.get_icurve()
            ax.plot(icurve.x, icurve.y, label=f'Component {comp.peak_index}')
        ax.set_xlabel('Frames')
        ax.set_ylabel('Intensity')
        ax.set_title('Components Elution Curves')
        ax.legend()
        fig.tight_layout()
        plt.show()

    ranges = []
    areas = []
    for comp in components:
        areas.append(comp.compute_area())
        ranges.append(comp.compute_range(area_ratio))

    area_proportions = np.array(areas)/np.sum(areas)
    if debug:
        print("area_proportions=", area_proportions)

    ret_ranges = []
    for comp, range_, prop, legacy_range in zip(components, ranges, area_proportions, legacy_ranges):
        minor = prop < MINOR_COMPONENT_MAX_PROP
        elm_recs = legacy_range[0].elm_recs     #  legacy_range[0] is a PeakInfo
        ret_ranges.append(comp.make_paired_range(range_, minor=minor, elm_recs=elm_recs, debug=debug))

    if debug:
        from importlib import reload
        import molass_legacy.Decomposer.UnifiedDecompResultTest
        reload(molass_legacy.Decomposer.UnifiedDecompResultTest)
        from molass_legacy.Decomposer.UnifiedDecompResultTest import plot_decomp_results
        editor_ranges = []
        for prange in ret_ranges:
            editor_ranges.append(prange.get_fromto_list())
        plot_decomp_results([decomp_result], editor_ranges)

    return ret_ranges


"""
Logging.Utils.py

Logging utilities for the Molass library.
"""

import logging
import sys

def setup_notebook_logging():
    """
    Setup logging to display messages in Jupyter notebooks.

    Returns
    -------
    logger : logging.Logger
        Configured logger for notebook output.
    """
    logger = logging.getLogger()  # Or use the specific logger name if needed
    logger.setLevel(logging.INFO)  # Or your desired level

    # Remove existing handlers to avoid duplicate logs
    logger.handlers = []

    # Add a StreamHandler for notebook output
    handler = logging.StreamHandler(sys.stdout)
    handler.setFormatter(logging.Formatter('%(levelname)s: %(message)s'))
    logger.addHandler(handler)
    return logger


"""
    LowRank.LowRankInfo.py

    This module contains the class LowRankInfo, which is used to store information
    about the components of a SecSaxsData, which is mathematically interpreted as
    a low rank approximation of a matrix.
"""
from importlib import reload
import numpy as np

class Decomposition:
    """
    A class to store the result of decomposition which is a low rank approximation.

    The result includes both components of X-ray and UV data and their associated information.

    Attributes
    ----------
    ssd : SecSaxsData
        The SecSaxsData object from which the decomposition was performed.
    xr : XrData
        The XrData object from the SecSaxsData.
    xr_icurve : Curve
        The i-curve used for the decomposition of the X-ray data.
    xr_ccurves : list of Curve
        The component curves for the X-ray data.
    xr_ranks : list of int or None
        The ranks for each component of the X-ray data. If None, default ranks are used.
    uv : UvData
        The UvData object from the SecSaxsData.
    uv_icurve : Curve
        The i-curve used for the decomposition of the UV data.
    uv_ccurves : list of Curve
        The component curves for the UV data.
    uv_ranks : list of int or None
        The ranks for each component of the UV data. If None, default ranks are used.
    mapping : MappingInfo
        The mapping information between the X-ray and UV data.
    mapped_curve : MappedCurve or None
        The mapped curve from the X-ray to UV domain. If None, it can be computed when needed.
    paired_ranges : list of PairedRange or None
        The paired ranges for the X-ray and UV data. If None, it can be computed when needed.
    num_components : int
        The number of components in the decomposition.
    """

    def __init__(self, ssd, xr_icurve, xr_ccurves, uv_icurve, uv_ccurves, mapped_curve=None, paired_ranges=None, **kwargs):
        """
        Initialize the Decomposition object.

        Parameters
        ----------
        ssd : SecSaxsData
            The SecSaxsData object from which the decomposition was performed.
        xr_icurve : Curve
            The i-curve used for the decomposition of the X-ray data.
        xr_ccurves : list of Curve
            The component curves for the X-ray data.
        uv_icurve : Curve
            The i-curve used for the decomposition of the UV data.
        uv_ccurves : list of Curve
            The component curves for the UV data.
        mapped_curve : MappedCurve, optional
            The mapped curve from the X-ray to UV domain. If None, it can be computed when needed.
        paired_ranges : list of PairedRange, optional
            The paired ranges for the X-ray and UV data. If None, it can be computed when needed.
        kwargs : dict, optional
            Additional keyword arguments (not used).
        """
        if uv_ccurves is None:
            pass
        else:
            assert len(xr_ccurves) == len(uv_ccurves)  
        self.num_components = len(xr_ccurves)
        self.ssd = ssd

        self.xr = ssd.xr
        self.xr_icurve = xr_icurve
        self.xr_ccurves = xr_ccurves
        self.xr_ranks = None

        self.uv = ssd.uv
        self.uv_icurve = uv_icurve
        self.uv_ccurves = uv_ccurves
        self.uv_ranks = None
 
        self.mapping = ssd.get_mapping()
        self.mapped_curve = mapped_curve
        self.paired_ranges = paired_ranges

        self.guinier_objects = None
        self.bounded_lrf_info = None
        self.model = xr_ccurves[0].model

    def copy_with_new_components(self, xr_ccurves, uv_ccurves):
        """
        Create a new Decomposition with new component curves.

        Parameters
        ----------
        xr_ccurves : list of Curve
            The new component curves for the X-ray data.
        uv_ccurves : list of Curve
            The new component curves for the UV data.

        Returns
        -------
        Decomposition
            A new Decomposition object with the specified component curves.
        """
        return Decomposition(self.ssd, self.xr_icurve, xr_ccurves, self.uv_icurve, uv_ccurves, self.mapped_curve, self.paired_ranges)

    @property
    def xr_components(self):
        """Alias for ``xr_ccurves`` — the XR component curves."""
        return self.xr_ccurves

    @property
    def uv_components(self):
        """Alias for ``uv_ccurves`` — the UV component curves."""
        return self.uv_ccurves

    def get_num_components(self):
        """
        Get the number of components.

        Returns
        -------
        int
            The number of components in the decomposition.
        """
        return self.num_components

    def get_guinier_objects(self, debug=False):
        """
        Get the list of Guinier objects for the XR components.

        Returns
        -------
        list of Guinier
            The list of Guinier objects for each XR component.
        """
        if self.guinier_objects is None:
            xr_components = self.get_xr_components(debug=debug)
            self.guinier_objects = [c.get_guinier_object() for c in xr_components]
        return self.guinier_objects

    def get_rgs(self):
        """
        Get the list of Rg values for the XR components.

        Returns
        -------
        rgs : list of float, length n_components
            Radius of gyration in **Ångströms (Å)** for each XR component,
            in the same order as ``get_xr_components()``.
            If Guinier fitting fails for a component, ``float('nan')`` is
            returned for that position (never ``None``), so the result is
            always safe to use in numeric / numpy operations.

            Guard pattern::

                import math
                for i, rg in enumerate(decomp.get_rgs()):
                    if math.isnan(rg):
                        print(f"Component {i+1}: Guinier fit failed")
                    else:
                        print(f"Component {i+1}: Rg = {rg:.2f} Å")
        """
        return [sv.Rg if sv.Rg is not None else float('nan')
                for sv in self.get_guinier_objects()]

    def get_rg_curve(self):
        """Compute the per-frame Rg curve from the raw XR data.

        Runs a Guinier fit on every elution frame independently and returns
        the results as an ``RgCurve`` object.  This is useful for assessing
        whether a peak is a pure single-component species (flat Rg vs. frame)
        or a heterogeneous mixture (varying Rg).

        .. note::
            This can be slow for large datasets because it fits one Guinier
            region per frame.

        Returns
        -------
        rgcurve : molass.Guinier.RgCurve.RgCurve
            An ``RgCurve`` with attributes:

            - ``.x`` — frame indices (integer array)
            - ``.y`` — Rg values in Å; ``NaN`` where Guinier fit failed
            - ``.scores`` — Guinier fit quality scores (0–1)

        Examples
        --------
        ::

            rgcurve = decomp.get_rg_curve()
            import matplotlib.pyplot as plt
            plt.plot(rgcurve.x, rgcurve.y, '.')
            plt.xlabel("Frame")
            plt.ylabel("Rg (Å)")
            plt.title("Rg vs. elution frame")
            plt.show()
        """
        return self.xr.compute_rgcurve()

    def get_P_at(self, q_target, normalize=False):
        """Return the XR scattering matrix P interpolated onto *q_target*.

        Parameters
        ----------
        q_target : array-like, shape (m,)
            Target q-values in Å⁻¹.
        normalize : bool, optional
            If ``True``, each component column is divided by its maximum so
            that all columns peak at 1.  Default ``False``.

        Returns
        -------
        P_interp : np.ndarray, shape (m, n_components)
            Scattering matrix P evaluated at *q_target*.
        """
        from molass.LowRank.AlignDecompositions import get_P_at
        return get_P_at(self, q_target, normalize=normalize)

    def component_quality_scores(self):
        """
        Compute a per-component reliability score in [0, 1].

        Scores blend Rg distinctiveness (70 %) and area proportion (30 %).
        A score of 0.0 means Guinier fitting failed or Rg is indistinguishable
        from another component's.  A score near 1.0 means the component is
        well-separated in Rg and carries a non-trivial fraction of the signal.

        Returns
        -------
        scores : list of float
            Reliability score for each component, in the same order as
            ``get_rgs()`` and ``get_proportions()``.

        See Also
        --------
        is_component_reliable : threshold-based boolean version.

        Examples
        --------
        ::

            scores = decomp.component_quality_scores()
            for i, s in enumerate(scores):
                print(f"Component {i+1}: reliability = {s:.2f}")
        """
        from molass.LowRank.ComponentReliability import component_quality_scores
        return component_quality_scores(self)

    def is_component_reliable(self, index, threshold=0.5):
        """
        Return ``True`` if component *index* has a quality score ≥ *threshold*.

        Parameters
        ----------
        index : int
            Zero-based component index.
        threshold : float, optional
            Minimum score considered reliable.  Default 0.5.

        Returns
        -------
        bool

        Examples
        --------
        ::

            if not decomp.is_component_reliable(1):
                print("Component 2 may be a noise artifact.")
        """
        from molass.LowRank.ComponentReliability import is_component_reliable
        return is_component_reliable(self, index, threshold=threshold)

    def plot_components(self, **kwargs):
        """decomposition.plot_components(title=None, fig=None, axes=None, **kwargs)

        Plot the components.

        Parameters
        ----------
        title : str, optional
            If specified, add a super title to the plot.
        fig : matplotlib.figure.Figure, optional
            An existing Figure to draw into. If ``None`` (default), a new
            figure is created automatically.
        axes : array-like of shape (2, 3), optional
            A 2×3 array of Axes to draw into. Must be provided together with
            ``fig`` when injecting into an existing subplot grid. If ``None``
            (default), axes are created automatically inside *fig*.

            Expected layout::

                axes[0, 0]  UV elution curves
                axes[0, 1]  UV absorbance curves
                axes[0, 2]  (UV spare / unused)
                axes[1, 0]  XR elution curves
                axes[1, 1]  XR scattering curves (log scale)
                axes[1, 2]  XR Guinier plot  ← Kratky is omitted when axes are injected

            .. note::
                When both *fig* and *axes* are provided the caller is
                responsible for creating axes with compatible geometry.

        Returns
        -------
        result : PlotResult
            A PlotResult object which contains the following attributes.
            
            - fig: The matplotlib Figure object.
            - axes: A 2×3 array of Axes objects.
        """
        debug = kwargs.get('debug', False)
        if debug:
            from importlib import reload
            import molass.PlotUtils.DecompositionPlot
            reload(molass.PlotUtils.DecompositionPlot)
        from molass.PlotUtils.DecompositionPlot import plot_components_impl, ALLOWED_KEYS
        for key in kwargs.keys():
            if key not in ALLOWED_KEYS:
                raise ValueError(f"Invalid key: {key}. Allowed keys are: {ALLOWED_KEYS}")
        return plot_components_impl(self, **kwargs)

    def update_xr_ranks(self, ranks, debug=False):
        """
        Update the ranks for the X-ray data.

        Default ranks are one for each component which means that interparticle interactions are not considered.
        This method allows the user to set different ranks for each component.

        Parameters
        ----------
        ranks : list of int
            The ranks for each component.

        Returns
        -------
        None
        """
        self.xr_ranks = ranks

    def get_xr_matrices(self, debug=False):
        """
        Get the factorized matrices for the X-ray (SAXS) data.

        Parameters
        ----------
        debug : bool, optional
            If True, enable debug mode.

        Returns
        -------
        M : np.ndarray, shape (n_q, n_frames)
            Measured scattering intensity matrix.
            Rows are q-points; columns are elution frames.
        C : np.ndarray, shape (n_components, n_frames)
            Elution curves (concentration profiles) for each component.
            Each row is one component's elution curve over frames.
        P : np.ndarray, shape (n_q, n_components)
            Scattering profiles (form factors) for each component.
            Each **column** is one component's P(q) in absolute or
            relative intensity units (matching the scale of M).
        Pe : np.ndarray, shape (n_q, n_components)
            Estimated error (standard deviation) on P,
            propagated from the measurement error matrix.

        Notes
        -----
        The q-values corresponding to the n_q rows are stored in
        ``decomp.xr.qv`` (shape ``(n_q,)``, units Å⁻¹).

        Example
        -------
        ::

            M, C, P, Pe = decomp.get_xr_matrices()
            # P[:, 0]  →  scattering profile of component 1
            # C[0, :]  →  elution curve of component 1
        """
        if debug:
            from importlib import reload
            import molass.LowRank.LowRankInfo
            reload(molass.LowRank.LowRankInfo)
        from molass.LowRank.LowRankInfo import compute_lowrank_matrices

        xr = self.xr

        # Step 1: naïve factorization (compute_lowrank_matrices unchanged)
        M_, C_, P_, Pe = compute_lowrank_matrices(
            xr.M, self.xr_ccurves, xr.E, self.xr_ranks, debug=debug)

        # Step 2: apply bounded LRF if any component is rank-2
        ranks = self.xr_ranks or [1] * self.num_components
        has_rank2 = any(r == 2 for r in ranks)

        if has_rank2:
            from molass.Guinier.RgEstimator import RgEstimator

            # Step 2a: pre-populate guinier_objects from naïve P
            if self.guinier_objects is None:
                guinier_objects = []
                for i in range(self.num_components):
                    jcurve_array = np.array([xr.qv, P_[:, i], Pe[:, i]]).T
                    guinier_objects.append(RgEstimator(jcurve_array))
                self.guinier_objects = guinier_objects

            # Step 2b: reconstruct full C and P (including c² rows/B columns)
            cy_list = [c.get_xy()[1] for c in self.xr_ccurves]
            for k, r in enumerate(ranks):
                if r > 1:
                    cy_list.append(cy_list[k] ** r)
            C_full = np.array(cy_list)
            P_full = M_ @ np.linalg.pinv(C_full)

            # Step 2c: apply bounded LRF
            if debug:
                from importlib import reload
                import molass.LowRank.BoundedLrf
                reload(molass.LowRank.BoundedLrf)
            from molass.LowRank.BoundedLrf import apply_bounded_lrf
            P_, bounded_info = apply_bounded_lrf(
                xr.qv, P_full, C_full, ranks, self.guinier_objects)
            self.bounded_lrf_info = bounded_info

            # Step 2d: re-propagate error for corrected P
            if xr.E is not None:
                from molass.LowRank.ErrorPropagate import compute_propagated_error
                Pe = compute_propagated_error(M_, P_, xr.E)

        return M_, C_, P_, Pe

    def get_xr_components(self, debug=False):
        """
        Get the per-component objects for the X-ray (SAXS) data.

        Parameters
        ----------
        debug : bool, optional
            If True, enable debug mode.

        Returns
        -------
        components : list of :class:`~molass.LowRank.Component.XrComponent`, length n_components
            One ``XrComponent`` per decomposed component, in component order.

            Each ``XrComponent`` exposes:

            - ``get_guinier_object()`` → Guinier fit result (Rg, I0, fit range)
            - ``get_jcurve_array()`` → ``np.ndarray`` shape ``(n_q, 3)``:
              columns are ``[q, P(q), Pe(q)]`` in Å⁻¹ and intensity units.
            - ``icurve_array`` → ``np.ndarray`` shape ``(2, n_frames)``:
              rows are ``[frame_x, elution_y]``.
            - ``compute_area()`` → scalar, integrated elution area.
        """
        if debug:
            from importlib import reload
            import molass.LowRank.Component
            reload(molass.LowRank.Component)
        from molass.LowRank.Component import XrComponent

        xr_matrices = self.get_xr_matrices(debug=debug)
        xrC, xrP, xrPe = xr_matrices[1:]

        ret_components = []
        for i in range(self.num_components):
            icurve_array = np.array([self.xr_icurve.x, xrC[i,:]])
            jcurve_array = np.array([self.xr.qv, xrP[:,i], xrPe[:,i]]).T
            ccurve = self.xr_ccurves[i]
            ret_components.append(XrComponent(icurve_array, jcurve_array, ccurve))

        return ret_components

    def get_uv_matrices(self, debug=False):
        """
        Get the matrices for the UV data.

        Parameters
        ----------
        debug : bool, optional
            If True, enable debug mode.

        Returns
        -------
        tuple of (np.ndarray, np.ndarray, np.ndarray, np.ndarray)
            The matrices for the UV data.
        """
        if debug:
            from importlib import reload
            import molass.LowRank.LowRankInfo
            reload(molass.LowRank.LowRankInfo)
        from molass.LowRank.LowRankInfo import compute_lowrank_matrices

        uv = self.uv
        return compute_lowrank_matrices(uv.M, self.uv_ccurves, uv.E, self.uv_ranks, debug=debug)

    def get_uv_components(self, debug=False):
        """
        Get the components for the UV data.

        Returns
        -------
        List of UvComponent objects.
        """
        if debug:
            from importlib import reload
            import molass.LowRank.Component
            reload(molass.LowRank.Component)
        from molass.LowRank.Component import UvComponent

        uv_matrices = self.get_uv_matrices(debug=debug)
        uvC, uvP, uvPe = uv_matrices[1:]
        if uvPe is None:
            uvPe = np.zeros_like(uvP)

        ret_components = []
        for i in range(self.num_components):
            uv_elution = np.array([self.uv_icurve.x, uvC[i,:]])
            uv_spectral = np.array([self.uv.wv, uvP[:,i], uvPe[:,i]]).T
            ccurve = self.uv_ccurves[i]
            ret_components.append(UvComponent(uv_elution, uv_spectral, ccurve))

        return ret_components

    def get_pairedranges(self, mapped_curve=None, area_ratio=0.7, concentration_datatype=2, debug=False):
        """
        Get the paired ranges.

        Parameters
        ----------
        mapped_curve : MappedCurve, optional
            If specified, use this mapped curve instead of computing a new one.
        area_ratio : float, optional
            The area ratio for the range computation.
        concentration_datatype : int, optional
            The concentration datatype for the range computation.
        debug : bool, optional
            If True, enable debug mode.

        Returns
        -------
        list of PairedRange
            The list of :class:`~molass.LowRank.PairedRange` objects.
        """
        if self.paired_ranges is None:
            if debug:
                import molass.Reports.ReportRange
                reload(molass.Reports.ReportRange)
            from molass.Reports.ReportRange import make_v1report_ranges_impl
            if mapped_curve is None:
                if self.mapped_curve is None:
                    from molass.Backward.MappedCurve import make_mapped_curve
                    self.mapped_curve = make_mapped_curve(self.ssd, debug=debug)
                mapped_curve = self.mapped_curve
            self.paired_ranges = make_v1report_ranges_impl(self, self.ssd, mapped_curve, area_ratio, concentration_datatype, debug=debug)
        return self.paired_ranges

    def get_proportions(self):
        """
        Get the relative area fractions of the XR components.

        Returns
        -------
        proportions : np.ndarray, shape (n_components,)
            Normalised elution-curve area fraction for each component,
            summing to 1.0.  Values are in the range [0, 1].
            Proportional to the amount (concentration × volume) of each
            species in the SEC peak region.
        """
        n = self.get_num_components()
        props = np.zeros(n)
        for i, c in enumerate(self.get_xr_components()):
            props[i] = c.compute_area()
        return props/np.sum(props)

    def compute_scds(self, debug=False):
        """
        Get the list of SCDs (Score of Concentration Dependence) for the decomposition.

        Returns
        -------
        list of float
            The list of SCD values for each component.
        """
        if debug:
            import molass.Backward.RankEstimator
            reload(molass.Backward.RankEstimator)
        from molass.Backward.RankEstimator import compute_scds_impl
        return compute_scds_impl(self, debug=debug)
    
    def get_cd_color_info(self):
        """
        Get the color information for the concentration dependence.

        Returns
        -------
        peak_top_xes : list of float
            The list of peak top x values for each component.
        scd_colors : list of str
            The list of colors for each component based on their ranks.
        """
        if self.xr_ranks is None:
            import logging
            logging.warning("Decomposition.get_cd_color_info: xr_ranks is None, using default ranks.")
            ranks = [1] * self.num_components
        else:
            ranks = self.xr_ranks

        peak_top_xes = [ccurve.get_peak_top_x() for ccurve in self.xr_ccurves]
        scd_colors = ['green' if rank == 1 else 'red' for rank in ranks]
        return peak_top_xes, scd_colors
    
    def optimize_with_model(self, model_name, rgcurve=None, model_params=None, debug=False):
        """
        Optimize the decomposition with a model.

        Parameters
        ----------
        model_name : str
            The name of the model to use for optimization.

            Supported models:

            - ``SDM``: `Stochastic Dispersive Model <https://biosaxs-dev.github.io/molass-essence/chapters/60/stochastic-theory.html#stochastic-dispersive-model>`_
            - ``EDM``: `Equilibrium Dispersive Model <https://biosaxs-dev.github.io/molass-essence/chapters/60/kinetic-theory.html#equilibrium-dispersive-model>`_

        rgcurve : Curve, optional
            The Rg curve to use for the optimization.

        model_params : dict, optional
            The parameters for the model.

        debug : bool, optional
            If True, enable debug mode.

        Returns
        -------
        result : Decomposition
            A new Decomposition object with optimized components.
        """
        if debug:
            import molass.SEC.ModelFactory
            reload(molass.SEC.ModelFactory)
        from molass.SEC.ModelFactory import create_model
        model = create_model(model_name, debug=debug)
        return model.optimize_decomposition(self, rgcurve=rgcurve, model_params=model_params, debug=debug)
    
    def make_rigorous_initparams(self, baseparams, debug=False):
        """
        Make initial parameters for rigorous optimization.

        Parameters
        ----------
        debug : bool, optional
            If True, enable debug mode.

        Returns
        -------
        np.ndarray
            The initial parameters for rigorous optimization.
        """
        if self.model == 'egh':
            if debug:
                import molass.Rigorous.RigorousEghParams
                reload(molass.Rigorous.RigorousEghParams)
            from molass.Rigorous.RigorousEghParams import make_rigorous_initparams_impl
            return make_rigorous_initparams_impl(self, baseparams, debug=debug)
        elif self.model == 'sdm':
            if debug:
                import molass.Rigorous.RigorousSdmParams
                reload(molass.Rigorous.RigorousSdmParams)
            from molass.Rigorous.RigorousSdmParams import make_rigorous_initparams_impl
            return make_rigorous_initparams_impl(self, baseparams, debug=debug)
        elif self.model == 'edm':
            if debug:
                import molass.Rigorous.RigorousEdmParams
                reload(molass.Rigorous.RigorousEdmParams)
            from molass.Rigorous.RigorousEdmParams import make_rigorous_initparams_impl
            return make_rigorous_initparams_impl(self, baseparams, debug=debug)
        else:
            raise ValueError(f"Decomposition.make_rigorous_initparams: Unsupported model '{self.model}'")

    def optimize_rigorously(self, rgcurve=None, analysis_folder=None, method='BH', niter=20, debug=False):
        """
        Perform a rigorous decomposition.

        Parameters
        ----------
        rgcurve : Curve
            The Rg curve to use for the decomposition.
        analysis_folder : str, optional
            The folder to save analysis results.
        method : str, optional
            The method to use for rigorous optimization. Default is 'BH'.
        niter : int, optional
            The number of iterations for the optimization. Default is 20.
        debug : bool, optional
            If True, enable debug mode.

        Returns
        -------
        result : Decomposition
            A new Decomposition object after rigorous decomposition.
        """
        if debug:
            import molass.Rigorous.RigorousImplement
            reload(molass.Rigorous.RigorousImplement)
        from molass.Rigorous.RigorousImplement import make_rigorous_decomposition_impl

        if rgcurve is None:
            rgcurve = self.ssd.xr.compute_rgcurve()

        return make_rigorous_decomposition_impl(self, rgcurve, analysis_folder=analysis_folder, method=method, niter=niter, debug=debug)



"""
    DenssUtils.py

    Copyright (c) 2019-2025, SAXS Team, KEK-PF

    Note that the original DENSS by T. Grant is licensed under GPL3.

    DENSS is available open source under the GPL 3.0 license.
    This means that anyone is free to download, utilize and develop the code,
    provided that any derivative code developed based on DENSS
    or its underlying algorithm is also distributed under the GPL 3.0 license
    and retains the original copyright.
"""
import numpy as np
import sys, argparse, os
import logging
import time
import molass.SAXS.denss as denss
import molass.SAXS.denss.options as dopts
from molass_legacy.KekLib.BasicUtils import Struct

MAXNUM_STEPS = 20000

def fit_data_impl(q, a, e, file=None, D=None, alpha=None, max_alpha=None, nes=2, extrapolate=True, gui=False, use_memory_data=False):
    """ fit_data_impl(q, a, e, file=None, D=None, alpha=None, max_alpha=None, nes=2, extrapolate=True, gui=False, use_memory_data=False)
    Fit the provided SAXS data using DENSS's fit_data logic.
    This function processes the input data and prepares it for DENSS reconstruction.

    Parameters
    ----------
    q : np.ndarray
        The q values of the SAXS data.
    a : np.ndarray
        The intensity values of the SAXS data.
    e : np.ndarray
        The error values of the SAXS data.
    file : str, optional
        The path to the data file. If None, data is taken from memory.
    D : float, optional
        The maximum dimension (Dmax) of the particle. If None, it will be estimated
        from the data.
    alpha : float, optional
        The regularization parameter. If None, it will be optimized.
    max_alpha : float, optional
        The maximum value for alpha during optimization.
    nes : int, optional
        The number of electrons per voxel.
    extrapolate : bool, optional
        If True, extrapolate the data to higher q values.
    gui : bool, optional
        If True, enable GUI mode for optimization.
    use_memory_data : bool, optional
        If True, use the provided q, a, e arrays instead of reading from file.

    Returns
    -------
    sasrec : denss.Sasrec
        The fitted Sasrec object containing the processed data.
    work_info : Struct
        A structure containing additional information about the fitting process.
    """

    # task: update this construction automatically from bin\denss.fit_data.py
    args = Struct(
        alpha = alpha,
        units = "a",
        max_alpha = max_alpha,
        output = "data_name",
        file = file,    # 
        dmax = D,
        nes = nes,
        n1 = None,
        n2 = None,
        ignore_errors = False,
        q = None,
        qmax = None,
        nq = None,
        r = None,
        max_dmax = None,
        qfile = None,
        rfile = None,
        extrapolate = extrapolate,
        nr = None,
        plot = True,
        log = True,
        write_shannon = True,
        )

    """
    DENSS strongly recommends to include I(q=0).
    To conform, we are using denss.fit_data.py here processing logic
    to modify the input data

    below is the extracted default processing from denss.fit_data.py
    lines[73:194] as of ver. 1.8.6
    """
### DENSS.bin.denss.fit_data.py copy & modify BEIGN ###
    alpha = args.alpha

    if args.output is None:
        fname_nopath = os.path.basename(args.file)
        basename, ext = os.path.splitext(fname_nopath)
        output = basename
    else:
        output = args.output

    if use_memory_data:
        # as invoked from LRF preview 
        Iq = np.array([q, a, e]).T
    else:
        if args.ignore_errors:
            Iq = np.genfromtxt(args.file, invalid_raise=False, usecols=(0, 1))
        else:
            Iq = np.genfromtxt(args.file, invalid_raise=False, usecols=(0, 1, 2))
    if len(Iq.shape) < 2:
        print("Invalid data format. Data file must have 3 columns: q, I, errors. Alternatively, disable errors with --ignore_errors option (sets errors to 1.0).")
        exit()
    if Iq.shape[1] < 3 or args.ignore_errors:
        print("WARNING: Only 2 columns given. Data should have 3 columns: q, I, errors.")
        print("WARNING: Setting error bars to 1.0 (i.e., ignoring error bars)")
        Iq2 = np.zeros((Iq.shape[0], 3))
        Iq2[:,:2] = Iq[:,:2]
        Iq2[:,2] += 1.0 #set error bars to 1.0
        Iq = Iq2
    Iq = Iq[~np.isnan(Iq).any(axis = 1)]
    #get rid of any data points equal to zero in the intensities or errors columns
    idx = np.where((Iq[:,1]!=0)&(Iq[:,2]!=0))
    Iq = Iq[idx]

    nes = args.nes

    if args.units == "nm":
        Iq[:,0] *= 0.1

    if args.n1 is None:
        n1 = 0
    else:
        n1 = args.n1
    if args.n2 is None:
        n2 = len(Iq[:,0])
    else:
        n2 = args.n2

    Iq_orig = np.copy(Iq)

    if args.dmax is None:
        #estimate dmax directly from data
        #note that denss.estimate_dmax does NOT extrapolate
        #the high q data, even though by default
        #denss.Sasrec does extrapolate.
        D, sasrec = denss.estimate_dmax(Iq, clean_up=True)
    else:
        D = args.dmax

    if args.max_dmax is None:
        args.max_dmax = 2.*D

    if args.rfile is not None:
        r = np.loadtxt(args.rfile,usecols=(0,))
    else:
        r = None

    if args.qfile is not None:
        qc = np.loadtxt(args.qfile,usecols=(0,))
    else:
        qc = None

    # print("Dmax = %.2f"%D)
    qmax = Iq[n1:n2,0].max()
    if (qc is None) and (args.extrapolate):
        qmaxc = qmax*3.0
    elif (qc is None):
        qmaxc = qmax
    else:
        qmaxc = qc.max()

    #let a user set a desired set of q values to be calculated
    #based on a given qmax and nq
    if (args.qmax is not None) or (args.nq is not None):
        if args.qmax is not None:
            qmaxc = args.qmax
        else:
            qmaxc = qmaxc
        if args.nq is not None:
            nqc = args.nq
        else:
            nqc = 501
        qc = np.linspace(0,qmaxc,nqc)
        #assume if they are giving args.qmax or args.nq, that they want to
        #disable extrapolation
        args.extrapolate = False

    nsh = qmax/(np.pi/D)
    nshc = qmaxc/(np.pi/D)
    # print("Number of experimental Shannon channels: %d"%(nsh))
    # print("Number of calculated Shannon channels: %d"%(nshc))
    if (nsh > 500) or (nshc>500):
        print("WARNING: Nsh > 500. Calculation may take a while. Please double check Dmax is accurate.")
        #give the user a few seconds to cancel with CTRL-C
        waittime = 10
        try:
            for i in range(waittime+1):
                sys.stdout.write("\rTo cancel, press CTRL-C in the next %d seconds. "%(waittime-i))
                sys.stdout.flush()
                time.sleep(1)
            print()
        except KeyboardInterrupt:
            print("Canceling...")
            exit()


    #calculate chi2 when alpha=0, to get the best possible chi2 for reference
    sasrec = denss.Sasrec(Iq[n1:n2], D, qc=qc, r=r, nr=args.nr, ne=nes, alpha=0.0, extrapolate=args.extrapolate)
    ideal_chi2 = sasrec.calc_chi2()

    if args.alpha is None:
        alpha = sasrec.optimize_alpha(gui=gui)
    else:
        alpha = args.alpha
    sasrec = denss.Sasrec(Iq[n1:n2], D, qc=qc, r=r, nr=args.nr, alpha=alpha, ne=nes, extrapolate=args.extrapolate)

    #implement method of estimating Vp, Vc, etc using oversmoothing
    sasrec.estimate_Vp_etal()

### DENSS.bin.denss.fit_data.py copy  & modify  END ###

    work_info = Struct(Iq=Iq, args=args, alpha=alpha, n1=n1, n2=n2, Iq_orig=Iq_orig)
    return sasrec, work_info

def fit_data(q, a, e, D=None, extrapolate=False, return_sasrec=False):
    """ fit_data(q, a, e, D=None, extrapolate=False, return_sasrec=False)
    Fit the provided SAXS data using DENSS's fit_data logic.
    This function processes the input data and prepares it for DENSS reconstruction.
    This is a simplified wrapper around fit_data_impl().

    Parameters
    ----------
    q : np.ndarray
        The q values of the SAXS data.
    a : np.ndarray
        The intensity values of the SAXS data.
    e : np.ndarray
        The error values of the SAXS data.
    D : float, optional
        The maximum dimension (Dmax) of the particle. If None, it will be estimated
        from the data.
    extrapolate : bool, optional
        If True, extrapolate the data to higher q values.
    return_sasrec : bool, optional
        If True, return the Sasrec object. If False, return qc, Ic, Icerr, D.

    Returns
    -------
    If return_sasrec is True:
        sasrec : denss.Sasrec
            The fitted Sasrec object containing the processed data.
    Else:
        qc : np.ndarray
            The q values used in the fitted data.
        Ic : np.ndarray
            The fitted intensity values.
        Icerr : np.ndarray
            The fitted error values.
        D : float
            The estimated maximum dimension (Dmax) of the particle.
    """
    sasrec, work_info = fit_data_impl(q, a, e,
                                        D=D,
                                        alpha=0,        # backward compatibility
                                        max_alpha=None,
                                        extrapolate=extrapolate,
                                        use_memory_data=True)
    if return_sasrec:
        return sasrec 
    else:
        return sasrec.qc, sasrec.Ic, sasrec.Icerr, sasrec.D

def fit_data_bc(q, a, e, extrapolate=False):    # backward compatible
    """ fit_data_bc(q, a, e, extrapolate=False)
    Fit the provided SAXS data using DENSS's fit_data logic.
    This function processes the input data and prepares it for DENSS reconstruction.
    This is a simplified wrapper around fit_data_impl().
    This function is for backward compatibility.
    It always sets alpha=0 and does not return the sasrec object.
    This function is for backward compatibility with older versions of molass.
    It is recommended to use fit_data() instead.

    Parameters
    ----------
    q : np.ndarray
        The q values of the SAXS data.
    a : np.ndarray
        The intensity values of the SAXS data.
    e : np.ndarray
        The error values of the SAXS data.
    extrapolate : bool, optional
        If True, extrapolate the data to higher q values.

    Returns
    -------
    qc : np.ndarray
        The q values used in the fitted data.
    Ic : np.ndarray
        The fitted intensity values.
    Icerr : np.ndarray
        The fitted error values.
    D : float
        The estimated maximum dimension (Dmax) of the particle.
    """
    sasrec, work_info = fit_data_impl(q, a, e,
                                        D=100,
                                        alpha=0,
                                        max_alpha=10,
                                        extrapolate=extrapolate,
                                        use_memory_data=True)
    return sasrec.qc, sasrec.Ic, sasrec.Icerr, sasrec.D

def get_dmax_with_datgnom(file_path):
    """ get_dmax_with_datgnom(file_path)
    Estimate Dmax from a DATGNOM file using AtsasDatGnomDdf.
    This function reads the DATGNOM file and estimates the best Dmax value.

    Parameters
    ----------
    file_path : str
        The path to the DATGNOM file.

    Returns
    -------
    dmax : float
        The estimated Dmax value.
    """
    from molass_legacy.Saxs.Rdf import AtsasDatGnomDdf
    ddf = AtsasDatGnomDdf(file_path)
    dmax = ddf.guess_best_dmax()
    return dmax

def get_argparser(return_args=False):
    """ get_argparser(return_args=False)
    Get the argument parser for DENSS.
    This function returns the argument parser for DENSS.
    If return_args is True, it returns the parsed arguments instead.
    This function is used to parse command line arguments for DENSS.

    Parameters
    ----------
    return_args : bool, optional
        If True, return the parsed arguments instead of the parser.

    Returns
    -------
    argparse.ArgumentParser or argparse.Namespace
        The argument parser or the parsed arguments.
    """
    parser = argparse.ArgumentParser(description="DENSS: DENsity from Solution Scattering.\n A tool for calculating an electron density map from solution scattering data",
                                     formatter_class=argparse.RawTextHelpFormatter)
    if return_args:
        args = dopts.parse_arguments(parser, return_args=True)
        return args
    else:
        return parser

def run_denss_impl(qc, ac, ec, dmax, data_name, steps=MAXNUM_STEPS, progress_cb=None, use_gpu=False, gui=False):
    """ run_denss_impl(qc, ac, ec, dmax, data_name, steps=MAXNUM_STEPS, progress_cb=None, use_gpu=False, gui=False)
    Run the DENSS reconstruction using the provided SAXS data.
    This function sets up and runs the DENSS reconstruction process.

    Parameters
    ----------
    qc : np.ndarray
        The q values of the SAXS data.
    ac : np.ndarray
        The intensity values of the SAXS data.
    ec : np.ndarray
        The error values of the SAXS data.
    dmax : float
        The maximum dimension (Dmax) of the particle.
    data_name : str
        The name of the data set.
    steps : int, optional
        The maximum number of steps for the reconstruction.
    progress_cb : callable, optional
        A callback function to report progress.
    use_gpu : bool, optional
        If True, use GPU acceleration for the reconstruction.
    gui : bool, optional
        If True, enable GUI mode for the reconstruction.

    Returns
    -------
    None
    """
    q = qc
    I = ac
    sigq = ec
    isfit = True    # it is assumed that the data is already fitted with fit_data_impl()

    sys.argv = ['dummy-script', '-f', data_name]
    if use_gpu:
        sys.argv += ['-gpu']
    parser = get_argparser()
    data_proxy = [q, I, sigq, I, dmax, isfit]
    args = dopts.parse_arguments(parser, data_proxy=data_proxy)

    qdata, Idata, sigqdata, qbinsc, Imean, chis, rg, supportV, rho, side, fit, final_chi2 = denss.reconstruct_abinitio_from_scattering_profile(
        # copied from denss/scripts/denss_abintio.py BEGIN
        q=args.q,
        I=args.I,
        sigq=args.sigq,
        dmax=args.dmax,
        qraw=args.qraw,
        Iraw=args.Iraw,
        sigqraw=args.sigqraw,
        ne=args.ne,
        voxel=args.voxel,
        oversampling=args.oversampling,
        recenter=args.recenter,
        recenter_steps=args.recenter_steps,
        recenter_mode=args.recenter_mode,
        positivity=args.positivity,
        positivity_steps=args.positivity_steps,
        extrapolate=args.extrapolate,
        output=args.output,
        steps=args.steps,
        ncs=args.ncs,
        ncs_steps=args.ncs_steps,
        ncs_axis=args.ncs_axis,
        ncs_type=args.ncs_type,
        seed=args.seed,
        support_start=args.support_start,
        shrinkwrap=args.shrinkwrap,
        shrinkwrap_old_method=args.shrinkwrap_old_method,
        shrinkwrap_sigma_start=args.shrinkwrap_sigma_start,
        shrinkwrap_sigma_end=args.shrinkwrap_sigma_end,
        shrinkwrap_sigma_decay=args.shrinkwrap_sigma_decay,
        shrinkwrap_threshold_fraction=args.shrinkwrap_threshold_fraction,
        shrinkwrap_iter=args.shrinkwrap_iter,
        shrinkwrap_minstep=args.shrinkwrap_minstep,
        chi_end_fraction=args.chi_end_fraction,
        write_xplor_format=args.write_xplor_format,
        write_freq=args.write_freq,
        enforce_connectivity=args.enforce_connectivity,
        enforce_connectivity_steps=args.enforce_connectivity_steps,
        enforce_connectivity_max_features=args.enforce_connectivity_max_features,
        cutout=args.cutout,
        quiet=args.quiet,
        # copied from denss/scripts/denss_abintio.py END
        gui=gui,
        DENSS_GPU=args.DENSS_GPU,
        progress_cb=progress_cb)

def run_denss_impl_dummy(qc, ac, ec, dmax, infile_name):
    """ run_denss_impl_dummy(qc, ac, ec, dmax, infile_name)
    A dummy implementation of run_denss_impl() for testing purposes.
    This function simulates the DENSS reconstruction process without performing any actual computation.
    It prints progress messages to the console.

    Parameters
    ----------
    qc : np.ndarray
        The q values of the SAXS data.
    ac : np.ndarray
        The intensity values of the SAXS data.
    ec : np.ndarray
        The error values of the SAXS data.
    dmax : float
        The maximum dimension (Dmax) of the particle.
    infile_name : str
        The name of the input data file.

    Returns
    -------

    """
    import time
    print("\n Step     Chi2     Rg    Support Volume")
    print(" ----- --------- ------- --------------")

    for j in range(100):
        time.sleep(0.1)
        sys.stdout.write("\r% 5i " % (j))
        sys.stdout.flush()

def get_denssfolder(parent=None):
    """ get_denssfolder(parent=None)
    Get the DENSS output folder.
    This function retrieves the DENSS output folder from the settings.
    If the folder does not exist, it creates it.

    Parameters
    ----------
    parent : object, optional
        The parent object to retrieve settings from. If None, uses global settings.

    Returns
    -------
    str or None
        The path to the DENSS output folder, or None if not set.
    """
    from molass_legacy._MOLASS.SerialSettings import get_setting
    from molass_legacy.KekLib.BasicUtils import mkdirs_with_retry

    analysis_folder = get_setting('analysis_folder')
    if analysis_folder is None or analysis_folder == "":
        return None

    denss_folder = (analysis_folder + '/DENSS').replace('\\', '/')
    if not os.path.exists(denss_folder):
        mkdirs_with_retry(denss_folder)

    return denss_folder

def get_outfolder(job_id=None, parent=None):
    """ get_outfolder(job_id=None, parent=None)
    Get the DENSS output folder for a specific job ID.
    This function retrieves the DENSS output folder and appends a job ID to it.
    If the folder does not exist, it creates it.

    Parameters
    ----------
    job_id : int, optional
        The job ID to append to the folder name. If None, uses '000' and increments if the folder exists.
    parent : object, optional
        The parent object to retrieve settings from. If None, uses global settings.

    Returns
    -------
    str or None
        The path to the DENSS output folder for the job, or None if the base folder
        is not set.
    """
    denss_folder = get_denssfolder(parent=parent)
    if denss_folder is None:
        return None

    import re
    if job_id is None:
        out_folder = denss_folder + '/000'
        while True:
            if os.path.exists(out_folder):
                out_folder = re.sub(r'/(\d+)$', lambda m: '/%03d' % (int(m.group(1)) + 1), out_folder)
            else:
                break
    else:
        out_folder_init = denss_folder + '/%03d' % job_id
        out_folder = out_folder_init
        i = 0
        while True:
            if os.path.exists(out_folder):
                i += 1
                out_folder =  out_folder_init + '_%d' % i
            else:
                break

    return out_folder

def get_denss_log_items(path):
    """ get_denss_log_items(path)
    Extract key-value pairs from a DENSS log file.
    This function reads a DENSS log file and extracts lines containing key-value pairs
    in the format "Final <key>: <value>".

    Parameters
    ----------
    path : str
        The path to the DENSS log file.

    Returns
    -------
    dict
        A dictionary containing the extracted key-value pairs.
    """
    import re
    item_re = re.compile(r'^\d+:\d+:\d+\s+Final\s+(.+):\s+(.+)$')
    ret_dict = {}
    with open(path) as log:
        for line in log:
            m = item_re.match(line)
            if m:
                ret_dict[m.group(1)] = float(m.group(2))
    return ret_dict

def run_pdb2mrc(in_file, queue=None):
    """ run_pdb2mrc(in_file, queue=None)
    Run the DENSS pdb2mrc script to convert a PDB file to an MRC file.
    This function invokes the DENSS pdb2mrc script as a subprocess to convert
    a PDB file to an MRC file. It can report progress through a queue.

    Parameters
    ----------
    in_file : str
        The path to the input PDB file.
    queue : multiprocessing.Queue, optional
        A queue to report progress. If None, no progress is reported.
        
    Returns
    -------
    str or None
        The path to the generated MRC file, or None if the conversion failed.
    """
    import time
    from molass_legacy.KekLib.SubProcess import Popen   # suppresses the child process window.
    from molass_legacy.KekLib.BasicUtils import get_home_folder
    print("Generating an mrc file.")
    script_path = os.path.join(get_home_folder(), r'molass\SAXS\denss\scripts\denss_pdb2mrc.py')
    python  = sys.executable.replace('pythonw.exe', 'python.exe')       # running with pythonw.exe seems inappropriate
    out_file = in_file.replace('.pdb', '')
    cmd = [python, script_path, '-f', in_file, '-o', out_file]
    # print(cmd)

    """
        currently not receiving the stdout of the child process.
        TODO: better print the progress.
    """
    p = Popen(cmd)
    while p.poll() is None:
        time.sleep(1)
        if queue is not None:
            queue.put((1, '.'))
    returncode = p.poll()

    if queue is not None:
        queue.put((1, '\n'))

    if returncode == 0:
        out_file = out_file + ".mrc"
        print("Generated %s" %  out_file)
    else:
        if queue is not None:
            queue.put((1, p.stderr.decode() ))
        out_file = None
        print("Failed with returncode: %d" % returncode)

    return out_file



"""
Peaks.RecognizerSpecific.py
"""

def bridge_recognize_peaks(x, y, num_peaks=None, debug=False):
    """
    Bridge function to recognize peaks in a curve using the legacy method.
    This function is a wrapper around the legacy peak recognition method
    when `num_peaks` is specified.
    Parameters
    ----------
    x : array-like
        The x-coordinates of the curve.
    y : array-like
        The y-coordinates of the curve.
    num_peaks : int, optional
        The number of peaks to recognize. If None, the legacy method will not be used.
    debug : bool, optional
        If True, additional debugging information will be printed and plotted.

    Returns
    -------
    list
        A list of indices where peaks are found in the curve.
    """
    from molass_legacy.QuickAnalysis.ModeledPeaks import recognize_peaks
    params_list = recognize_peaks(x, y, num_peaks=num_peaks, exact_num_peaks=num_peaks, debug=debug)
    peaks = []
    for h, m, s, t in params_list:
        peaks.append(int(round(m - x[0])))
    if debug:
        import matplotlib.pyplot as plt
        plt.plot(x, y, label='Curve')
        plt.scatter(x[peaks], y[peaks], color='red', label='Peaks')
        plt.legend()
        plt.show()
    return peaks


"""
    Mapping.RatioCurve.py
"""
import numpy as np
from molass.DataObjects.Curve import Curve

class RatioCurve(Curve):
    """
    Represents the ratio curve between mapped UV data and XR data.

    Attributes
    ----------
    x : array-like
        The x-values of the ratio curve.
    y : array-like
        The y-values of the ratio curve, representing the ratio of mapped UV intensity to XR intensity.
    """
    def __init__(self, x, y, type="i"):
        """
        Initializes the RatioCurve with given x and y values.

        Parameters
        ----------
        x : array-like
            The x-values of the ratio curve.
        y : array-like
            The y-values of the ratio curve.
        type : str, optional
            The type of the ratio curve (default is "i").
        """
        super().__init__(x, y, type=type)

def _compute_ratio_curve_impl(mapping, mp_curve=None, data_threshold=0.05, debug=False):
    if debug:
        import molass.DataUtils.Outliers
        from importlib import reload
        reload(molass.DataUtils.Outliers)
    from molass.DataUtils.Outliers import remove_outliers
    xr_curve = mapping.xr_curve
    if mp_curve is None:
        uv_curve = mapping.uv_curve
        mp_curve = mapping.get_mapped_curve(xr_curve, uv_curve, inverse_range=True, debug=debug)

    spline = xr_curve.get_spline()
    xr_y = spline(mp_curve.x)
    ratio = remove_outliers(mp_curve.y / xr_y, repeat=2)
    uv_valid = np.abs(mp_curve.y) > mp_curve.get_max_y()*data_threshold
    xr_valid = np.abs(xr_y) > xr_curve.get_max_y()*data_threshold
    valid = np.logical_and(uv_valid, xr_valid)
    if not np.any(valid):
        raise ValueError("No valid data points found for ratio calculation.")
    ratio[~valid] = np.nan  # Set invalid points to NaN
    return RatioCurve(mp_curve.x, ratio)